# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rangeWidgetUI.ui'
#
# Created by: PyQt5 UI code generator 5.15.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(1440, 817)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.frame_3 = QtWidgets.QFrame(Form)
        self.frame_3.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame_3.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_3.setObjectName("frame_3")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.frame_3)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.titleLabel = QtWidgets.QLabel(self.frame_3)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.titleLabel.setFont(font)
        self.titleLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.titleLabel.setObjectName("titleLabel")
        self.verticalLayout.addWidget(self.titleLabel)
        self.frame = QtWidgets.QFrame(self.frame_3)
        self.frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setLineWidth(0)
        self.frame.setObjectName("frame")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.frame)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.rangeLabel = QtWidgets.QLabel(self.frame)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.rangeLabel.setFont(font)
        self.rangeLabel.setObjectName("rangeLabel")
        self.horizontalLayout.addWidget(self.rangeLabel)
        self.low = QtWidgets.QDoubleSpinBox(self.frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.low.sizePolicy().hasHeightForWidth())
        self.low.setSizePolicy(sizePolicy)
        self.low.setMinimumSize(QtCore.QSize(0, 0))
        self.low.setMaximumSize(QtCore.QSize(68, 16777215))
        self.low.setMaximum(1000000.0)
        self.low.setSingleStep(0.1)
        self.low.setProperty("value", 0.0)
        self.low.setObjectName("low")
        self.horizontalLayout.addWidget(self.low)
        self.label_4 = QtWidgets.QLabel(self.frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_4.sizePolicy().hasHeightForWidth())
        self.label_4.setSizePolicy(sizePolicy)
        self.label_4.setMaximumSize(QtCore.QSize(20, 16777215))
        self.label_4.setAlignment(QtCore.Qt.AlignCenter)
        self.label_4.setObjectName("label_4")
        self.horizontalLayout.addWidget(self.label_4)
        self.high = QtWidgets.QDoubleSpinBox(self.frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.high.sizePolicy().hasHeightForWidth())
        self.high.setSizePolicy(sizePolicy)
        self.high.setMaximumSize(QtCore.QSize(68, 16777215))
        self.high.setSingleStep(0.1)
        self.high.setObjectName("high")
        self.horizontalLayout.addWidget(self.high)
        self.units = QtWidgets.QComboBox(self.frame)
        self.units.setMaximumSize(QtCore.QSize(68, 16777215))
        self.units.setObjectName("units")
        self.horizontalLayout.addWidget(self.units)
        self.verticalLayout.addWidget(self.frame)
        self.frame_2 = QtWidgets.QFrame(self.frame_3)
        self.frame_2.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_2.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_2.setLineWidth(0)
        self.frame_2.setObjectName("frame_2")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.frame_2)
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.pointsLabel = QtWidgets.QLabel(self.frame_2)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.pointsLabel.setFont(font)
        self.pointsLabel.setObjectName("pointsLabel")
        self.horizontalLayout_2.addWidget(self.pointsLabel)
        self.points = QtWidgets.QSpinBox(self.frame_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.points.sizePolicy().hasHeightForWidth())
        self.points.setSizePolicy(sizePolicy)
        self.points.setMaximumSize(QtCore.QSize(68, 16777215))
        self.points.setMinimum(1)
        self.points.setMaximum(10000000)
        self.points.setObjectName("points")
        self.horizontalLayout_2.addWidget(self.points)
        self.stepLabel = QtWidgets.QLabel(self.frame_2)
        self.stepLabel.setMaximumSize(QtCore.QSize(60, 16777215))
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.stepLabel.setFont(font)
        self.stepLabel.setObjectName("stepLabel")
        self.horizontalLayout_2.addWidget(self.stepLabel)
        self.step = QtWidgets.QDoubleSpinBox(self.frame_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.step.sizePolicy().hasHeightForWidth())
        self.step.setSizePolicy(sizePolicy)
        self.step.setMaximumSize(QtCore.QSize(68, 16777215))
        self.step.setDecimals(3)
        self.step.setMinimum(0.001)
        self.step.setMaximum(1000000.0)
        self.step.setSingleStep(0.1)
        self.step.setObjectName("step")
        self.horizontalLayout_2.addWidget(self.step)
        self.verticalLayout.addWidget(self.frame_2)
        self.verticalLayout_2.addWidget(self.frame_3)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.titleLabel.setText(_translate("Form", "Title"))
        self.rangeLabel.setText(_translate("Form", "Range"))
        self.label_4.setText(_translate("Form", "to"))
        self.pointsLabel.setText(_translate("Form", "Points"))
        self.stepLabel.setText(_translate("Form", "Step Size"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Form = QtWidgets.QWidget()
    ui = Ui_Form()
    ui.setupUi(Form)
    Form.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
"""
@author: thampy, walroth
"""

# Standard library imports
import os
import re
import time
import glob
import fnmatch
import numpy as np
from pathlib import Path
from collections import deque

# pyFAI imports
import fabio

# Qt imports
from pyqtgraph import Qt
from pyqtgraph.Qt import QtWidgets
from pyqtgraph.parametertree import ParameterTree, Parameter

# This module imports
from xdart.modules.ewald import EwaldArch, EwaldSphere
from .wrangler_widget import wranglerWidget, wranglerThread, wranglerProcess
from .ui.specUI import Ui_Form
from ....gui_utils import NamedActionParameter
from xdart.utils import get_img_data, get_img_meta
from xdart.utils import split_file_name, get_scan_name, get_img_number, get_fname_dir, get_sname_img_number
from xdart.utils import match_img_detector, get_series_avg, get_specFile, get_mask_array
from xdart.utils import write_xye, write_csv
from xdart.utils.containers.poni import get_poni_dict
# from xdart.utils import natural_sort_ints

from ....widgets import commandLine
from xdart.modules.pySSRL_bServer.bServer_funcs import specCommand

from icecream import ic; ic.configureOutput(prefix='', includeContext=True)

QFileDialog = QtWidgets.QFileDialog
QDialog = QtWidgets.QDialog
QMessageBox = QtWidgets.QMessageBox
QPushButton = QtWidgets.QPushButton

def_poni_file = '/Users/vthampy/SSRL_Data/RDA/static_det_test_data/test_xfc_data/test_xfc.poni'
def_img_file = '/Users/vthampy/SSRL_Data/RDA/static_det_test_data/test_xfc_data/images/images_0005.tif'

if not os.path.exists(def_poni_file):
    def_poni_file = ''
    def_img_file = ''

params = [
    {'name': 'Calibration', 'type': 'group', 'children': [
        {'name': 'poni_file', 'title': 'PONI File    ', 'type': 'str', 'value': def_poni_file},
        NamedActionParameter(name='poni_file_browse', title='Browse...'),
    ], 'expanded': True},
    {'name': 'Signal', 'type': 'group', 'children': [
        {'name': 'inp_type', 'title': '', 'type': 'list',
         'values': ['Image Series', 'Image Directory', 'Single Image'], 'value': 'Image Series'},
        {'name': 'File', 'title': 'Image File   ', 'type': 'str', 'value': def_img_file},
        NamedActionParameter(name='img_file_browse', title='Browse...'),
        {'name': 'img_dir', 'title': 'Directory', 'type': 'str', 'value': '', 'visible': False},
        NamedActionParameter(name='img_dir_browse', title='Browse...', visible=False),
        {'name': 'include_subdir', 'title': 'Subdirectories', 'type': 'bool', 'value': False, 'visible': False},
        {'name': 'img_ext', 'title': 'File Type  ', 'type': 'list',
         'values': ['tif', 'raw', 'h5', 'mar3450'], 'value': 'tif', 'visible': False},
        {'name': 'series_average', 'title': 'Average Scan', 'type': 'bool', 'value': False, 'visible': True},
        {'name': 'meta_ext', 'title': 'Meta File', 'type': 'list',
         'values': ['None', 'txt', 'pdi', 'SPEC'], 'value': 'pdi'},
        {'name': 'Filter', 'type': 'str', 'value': '', 'visible': False},
        {'name': 'write_mode', 'title': 'Write Mode  ', 'type': 'list',
         'values': ['Append', 'Overwrite'], 'value': 'Append'},
        {'name': 'mask_file', 'title': 'Mask File', 'type': 'str', 'value': ''},
        NamedActionParameter(name='mask_file_browse', title='Browse...'),
    ], 'expanded': True, 'visible': False},
    {'name': 'BG', 'title': 'Background', 'type': 'group', 'children': [
        {'name': 'bg_type', 'title': '', 'type': 'list',
         'values': ['None', 'Single BG File', 'Series Average', 'BG Directory'], 'value': 'None'},
        {'name': 'File', 'title': 'BG File', 'type': 'str', 'value': '', 'visible': False},
        NamedActionParameter(name='bg_file_browse', title='Browse...', visible=False),
        {'name': 'Match', 'title': 'Match Parameter', 'type': 'group', 'children': [
            {'name': 'Parameter', 'type': 'list', 'values': ['None'], 'value': 'None'},
            {'name': 'match_fname', 'title': 'Match File Root', 'type': 'bool', 'value': False},
            {'name': 'bg_dir', 'title': 'Directory', 'type': 'str', 'value': ''},
            NamedActionParameter(name='bg_dir_browse', title='Browse...'),
            {'name': 'Filter', 'type': 'str', 'value': ''},
        ], 'expanded': True, 'visible': False},
        {'name': 'Scale', 'type': 'float', 'value': 1, 'visible': False},
        {'name': 'norm_channel', 'title': 'Normalize', 'type': 'list', 'values': ['bstop'], 'value': 'bstop',
         'visible': False},
    ], 'expanded': False, 'visible': False},
    {'name': 'GI', 'title': 'Grazing Incidence', 'type': 'group', 'children': [
        {'name': 'Grazing', 'type': 'bool', 'value': False},
        {'name': 'th_motor', 'title': 'Theta Motor', 'type': 'list', 'values': ['th'], 'value': 'th'},
        {'name': 'th_val', 'title': 'Theta', 'type': 'str', 'value': '0.1', 'visible': False},
    ], 'expanded': False, 'visible': False},
    {'name': 'h5_dir', 'title': 'Save Path', 'type': 'str', 'value': get_fname_dir(), 'enabled': False},
    NamedActionParameter(name='h5_dir_browse', title='Browse...', visible=False),
    {'name': 'Timeout', 'type': 'float', 'value': 1, 'visible': False},
]

ctr = 1

class specWrangler(wranglerWidget):
    """Widget for integrating data associated with spec file. Can be
    used "live", will continue to poll data folders until image data
    and corresponding spec data are available.

    attributes:
        command_queue: Queue, used to send commands to thread
        file_lock, mp.Condition, process safe lock for file access
        fname: str, path to data file
        parameters: pyqtgraph Parameter, stores parameters from user
        scan_name: str, current scan name, used to handle syncing data
        sphere_args: dict, used as **kwargs in sphere initialization.
            see EwaldSphere.
        thread: wranglerThread or subclass, QThread for controlling
            processes
        timeout: int, how long before thread stops looking for new
            data.
        tree: pyqtgraph ParameterTree, stores and organizes parameters
        ui: Ui_Form from qtdesigner

    methods:
        cont, pause, stop: functions to pass continue, pause, and stop
            commands to thread via command_queue
        enabled: Enables or disables interactivity
        set_image_dir: sets the image directory
        set_poni_file: sets the calibration poni file
        set_spec_file: sets the spec data file
        set_fname: Method to safely change file name
        setup: Syncs thread parameters prior to starting

    signals:
        finished: Connected to thread.finished signal
        sigStart: Tells tthetaWidget to start the thread and prepare
            for new data.
        sigUpdateData: int, signals a new arch has been added.
        sigUpdateFile: (str, str, bool, str, bool, bool), sends new scan_name, file name
            GI flag (grazing incidence), theta motor for GI, single_image and
            series_average flag to static_scan_Widget.
        sigUpdateGI: bool, signals the grazing incidence condition has changed.
        showLabel: str, connected to thread showLabel signal, sets text
            in specLabel
    """
    showLabel = Qt.QtCore.Signal(str)

    def __init__(self, fname, file_lock, sphere, data_1d, data_2d, parent=None):
        """fname: str, file path
        file_lock: mp.Condition, process safe lock
        """
        super().__init__(fname, file_lock, parent)

        # Scan Parameters
        self.poni_dict = None
        self.scan_parameters = []
        self.counters = []
        self.motors = []
        self.command = None
        self.sphere = sphere
        self.data_1d = data_1d
        self.data_2d = data_2d

        # Setup gui elements
        self.ui = Ui_Form()
        self.ui.setupUi(self)
        self.ui.startButton.clicked.connect(self.start)
        # self.ui.startButton.clicked.connect(self.sigStart.emit)
        self.ui.pauseButton.clicked.connect(self.pause)
        self.ui.stopButton.clicked.connect(self.stop)
        self.ui.continueButton.clicked.connect(self.cont)

        # SpecCommand Line
        self.specCommandLine = commandLine(self)
        self.specCommandLine.send_command = self.send_command
        self.ui.commandLayout.addWidget(self.specCommandLine)
        self.buttonSend = QtWidgets.QPushButton(self)
        self.buttonSend.setText('Send')
        self.buttonSend.clicked.connect(self.send_command)
        self.ui.commandLayout.addWidget(self.buttonSend)
        self.commands = ['']
        self.current = -1
        self.keep_trying = True
        self.showLabel.connect(self.ui.specLabel.setText)

        # Setup parameter tree
        self.tree = ParameterTree()
        self.stylize_ParameterTree()
        self.parameters = Parameter.create(
            name='spec_wrangler', type='group', children=params
        )
        self.tree.setParameters(self.parameters, showTop=False)
        self.layout = Qt.QtWidgets.QVBoxLayout(self.ui.paramFrame)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.addWidget(self.tree)

        # Set attributes from Parameter Tree and a couple more
        # Calibration
        self.poni_file = self.parameters.child('Calibration').child('poni_file').value()
        self.get_poni_dict()

        # Signal
        self.inp_type = self.parameters.child('Signal').child('inp_type').value()
        self.img_file = self.parameters.child('Signal').child('File').value()
        self.img_dir = self.parameters.child('Signal').child('img_dir').value()
        self.include_subdir = self.parameters.child('Signal').child('include_subdir').value()
        self.img_ext = self.parameters.child('Signal').child('img_ext').value()
        self.single_img = True if self.inp_type == 'Single Image' else False
        self.file_filter = self.parameters.child('Signal').child('Filter').value()
        self.series_average = self.parameters.child('Signal').child('series_average').value()
        self.meta_ext = self.parameters.child('Signal').child('meta_ext').value()
        if self.meta_ext == 'None':
            self.meta_ext = None

        # Mask
        self.mask_file = self.parameters.child('Signal').child('mask_file').value()

        # Write Mode
        self.write_mode = self.parameters.child('Signal').child('write_mode').value()

        # Background
        self.bg_type = self.parameters.child('BG').child('bg_type').value()
        self.bg_file = self.parameters.child('BG').child('File').value()
        self.bg_dir = self.parameters.child('BG').child('Match').child('bg_dir').value()
        self.bg_matching_par = self.parameters.child('BG').child('Match').child('Parameter').value()
        self.bg_match_fname = self.parameters.child('BG').child('Match').child('match_fname').value()
        self.bg_file_filter = self.parameters.child('BG').child('Match').child('Filter').value()
        self.bg_scale = self.parameters.child('BG').child('Scale').value()
        self.bg_norm_channel = self.parameters.child('BG').child('norm_channel').value()

        # Grazing Incidence
        self.gi = self.parameters.child('GI').child('Grazing').value()
        self.th_mtr = self.parameters.child('GI').child('th_motor').value()

        # Timeout
        self.timeout = self.parameters.child('Timeout').value()

        # HDF5 Save Path
        self.h5_dir = self.parameters.child('h5_dir').value()

        # Wire signals from parameter tree based buttons
        self.parameters.sigTreeStateChanged.connect(self.setup)

        self.parameters.child('Calibration').child('poni_file_browse').sigActivated.connect(
            self.set_poni_file
        )
        self.parameters.child('Calibration').child('poni_file').sigValueChanged.connect(
            self.get_poni_dict
        )
        self.parameters.child('Signal').child('inp_type').sigValueChanged.connect(
            self.set_inp_type
        )
        self.parameters.child('Signal').child('img_file_browse').sigActivated.connect(
            self.set_img_file
        )
        self.parameters.child('Signal').child('img_dir_browse').sigActivated.connect(
            self.set_img_dir
        )
        self.parameters.child('Signal').child('mask_file_browse').sigActivated.connect(
            self.set_mask_file
        )
        self.parameters.child('Signal').child('series_average').sigValueChanged.connect(
            self.set_series_average
        )
        self.parameters.child('Signal').child('meta_ext').sigValueChanged.connect(
            self.set_meta_ext
        )
        self.parameters.child('BG').child('bg_type').sigValueChanged.connect(
            self.set_bg_type
        )
        self.parameters.child('BG').child('bg_file_browse').sigActivated.connect(
            self.set_bg_file
        )
        self.parameters.child('BG').child('Match').child('bg_dir_browse').sigActivated.connect(
            self.set_bg_dir
        )
        self.parameters.child('BG').child('Match').child('Parameter').sigValueChanged.connect(
            self.set_bg_matching_par
        )
        self.parameters.child('BG').child('norm_channel').sigValueChanged.connect(
            self.set_bg_norm_channel
        )
        self.parameters.child('GI').child('th_motor').sigValueChanged.connect(
            self.set_gi_th_motor
        )
        self.parameters.child('GI').child('th_val').sigValueChanged.connect(
            self.set_gi_th_motor
        )
        self.parameters.child('h5_dir_browse').sigActivated.connect(
            self.set_h5_dir
        )

        # Setup thread
        self.thread = specThread(
            self.command_queue,
            self.sphere_args,
            self.file_lock,
            self.fname,
            self.h5_dir,
            self.scan_name,
            self.single_img,
            self.poni_dict,
            self.inp_type,
            self.img_file,
            self.img_dir,
            self.include_subdir,
            self.img_ext,
            self.series_average,
            self.meta_ext,
            self.file_filter,
            self.mask_file,
            self.write_mode,
            self.bg_type,
            self.bg_file,
            self.bg_dir,
            self.bg_matching_par,
            self.bg_match_fname,
            self.bg_file_filter,
            self.bg_scale,
            self.bg_norm_channel,
            self.gi,
            self.th_mtr,
            self.timeout,
            self.command,
            self.sphere,
            self.data_1d,
            self.data_2d,
            self
        )

        self.thread.showLabel.connect(self.ui.specLabel.setText)
        self.thread.sigUpdateFile.connect(self.sigUpdateFile.emit)
        self.thread.finished.connect(self.finished.emit)
        self.thread.sigUpdate.connect(self.sigUpdateData.emit)
        # self.thread.sigUpdateArch.connect(self.sigUpdateArch.emit)
        self.thread.sigUpdateGI.connect(self.sigUpdateGI.emit)

        # Enable/disable buttons initially
        self.ui.pauseButton.setEnabled(False)
        self.ui.continueButton.setEnabled(False)
        self.ui.stopButton.setEnabled(False)

        self.setup()

    def setup(self):
        """Sets up the child thread, syncs all parameters.
        """
        # Calibration
        global ctr
        ctr += 1
        ic(ctr)
        if ctr > 50:
            exit()

        self.poni_file = self.parameters.child('Calibration').child('poni_file').value()
        self.thread.poni_dict = self.poni_dict

        # Signal
        self.file_filter = self.parameters.child('Signal').child('Filter').value()
        self.thread.file_filter = self.file_filter

        self.inp_type = self.parameters.child('Signal').child('inp_type').value()
        self.thread.inp_type = self.inp_type

        self.get_img_fname()
        self.thread.img_file = self.img_file

        self.scan_name = get_scan_name(self.img_file)
        self.thread.scan_name = self.scan_name

        self.thread.single_img = self.single_img
        self.thread.img_dir, self.thread.img_ext = self.img_dir, self.img_ext

        self.include_subdir = self.parameters.child('Signal').child('include_subdir').value()
        self.thread.include_subdir = self.include_subdir

        self.thread.series_average = self.series_average
        self.thread.meta_ext = self.meta_ext

        self.thread.h5_dir = self.h5_dir
        self.fname = os.path.join(self.h5_dir, self.scan_name + '.hdf5')
        self.thread.fname = self.fname

        self.mask_file = self.parameters.child('Signal').child('mask_file').value()
        self.thread.mask_file = self.mask_file

        # Write Mode
        self.write_mode = self.parameters.child('Signal').child('write_mode').value()
        self.thread.write_mode = self.write_mode

        # Background
        self.bg_type = self.parameters.child('BG').child('bg_type').value()
        self.thread.bg_type = self.bg_type

        self.bg_file = self.parameters.child('BG').child('File').value()
        self.thread.bg_file = self.bg_file

        self.bg_matching_par = self.parameters.child('BG').child('Match').child('Parameter').value()
        self.thread.bg_matching_par = self.bg_matching_par

        self.bg_dir = self.parameters.child('BG').child('Match').child('bg_dir').value()
        self.thread.bg_dir = self.bg_dir

        self.bg_match_fname = self.parameters.child('BG').child('Match').child('match_fname').value()
        self.thread.bg_match_fname = self.bg_match_fname

        self.bg_file_filter = self.parameters.child('BG').child('Match').child('Filter').value()
        self.thread.bg_file_filter = self.bg_file_filter

        self.bg_scale = self.parameters.child('BG').child('Scale').value()
        self.thread.bg_scale = self.bg_scale

        self.bg_norm_channel = self.parameters.child('BG').child('norm_channel').value()
        self.thread.bg_norm_channel = self.bg_norm_channel

        # Grazing Incidence
        self.gi = self.parameters.child('GI').child('Grazing').value()
        self.thread.gi = self.gi

        # self.th_mtr = self.parameters.child('GI').child('th_motor').value()
        self.thread.th_mtr = self.th_mtr
        # ic(self.th_mtr)

        # Timeout
        self.timeout = self.parameters.child('Timeout').value()
        self.thread.timeout = self.timeout

        self.thread.command = self.command

        self.thread.file_lock = self.file_lock
        self.thread.sphere_args = self.sphere_args

        self.thread.sphere = self.sphere
        self.thread.data_1d = self.data_1d
        self.thread.data_2d = self.data_2d

    def send_command(self):
        """Sends command in command line to spec, and calls
        commandLine send_command method to add command to list of
        commands.
        """
        command = self.specCommandLine.text()
        if not (command.isspace() or command == ''):
            try:
                specCommand(command, queue=True)
            except Exception as e:
                print(e)
                print(f"Command '{command}' not sent")

        commandLine.send_command(self.specCommandLine)

    def start(self):
        self.command = 'start'
        self.thread.command = 'start'
        self.ui.pauseButton.setEnabled(True)
        self.ui.continueButton.setEnabled(False)
        self.ui.stopButton.setEnabled(True)
        self.sigStart.emit()

    def pause(self):
        self.command = 'pause'
        self.thread.command = 'pause'
        # if self.thread.isRunning():
        self.ui.pauseButton.setEnabled(False)
        self.ui.continueButton.setEnabled(True)

    def cont(self):
        self.command = 'continue'
        self.thread.command = 'continue'
        # if self.thread.isRunning():
        self.ui.pauseButton.setEnabled(True)
        self.ui.continueButton.setEnabled(False)

    def stop(self):
        self.command = 'stop'
        self.thread.command = 'stop'
        # if self.thread.isRunning():
        self.ui.pauseButton.setEnabled(False)
        self.ui.continueButton.setEnabled(False)
        self.ui.stopButton.setEnabled(False)

    def set_poni_file(self):
        """Opens file dialogue and sets the calibration file
        """
        fname, _ = QFileDialog().getOpenFileName(
            filter="PONI (*.poni *.PONI)"
        )
        if fname != '':
            self.parameters.child('Calibration').child('poni_file').setValue(fname)
            self.poni_file = fname

    def get_poni_dict(self):
        """Opens file dialogue and sets the calibration file
        """
        if not os.path.exists(self.poni_file):
            for child in self.parameters.children():
                child.hide()
            self.parameters.child('Calibration').show()
            return

        self.poni_dict = get_poni_dict(self.poni_file)
        if self.poni_dict is None:
            print('Invalid Poni File')
            self.thread.signal_q.put(('message', 'Invalid Poni File'))
            return

        for child in self.parameters.children():
            child.show()

    def set_inp_type(self):
        """Change Parameter Names depending on Input Type
        """
        # ic()
        self.single_img = False
        self.parameters.child('Signal').child('File').show()
        self.parameters.child('Signal').child('img_file_browse').show()
        self.parameters.child('Signal').child('img_dir').hide()
        self.parameters.child('Signal').child('img_dir_browse').hide()
        self.parameters.child('Signal').child('include_subdir').hide()
        self.parameters.child('Signal').child('Filter').hide()
        self.parameters.child('Signal').child('series_average').show()
        self.parameters.child('Signal').child('img_ext').hide()

        inp_type = self.parameters.child('Signal').child('inp_type').value()
        if inp_type == 'Image Directory':
            self.parameters.child('Signal').child('File').hide()
            self.parameters.child('Signal').child('img_file_browse').hide()
            self.parameters.child('Signal').child('img_dir').show()
            self.parameters.child('Signal').child('img_dir_browse').show()
            self.parameters.child('Signal').child('include_subdir').show()
            self.parameters.child('Signal').child('Filter').show()
            self.parameters.child('Signal').child('img_ext').show()

        if inp_type == 'Single Image':
            self.single_img = True
            self.parameters.child('Signal').child('series_average').hide()

        self.inp_type = inp_type
        self.get_img_fname()

    def set_img_file(self):
        """Opens file dialogue and sets the spec data file
        """
        # ic()
        fname, _ = QFileDialog().getOpenFileName(
            filter="Images (*.tiff *.tif *.h5 *.raw *.mar3450)"
        )
        if fname != '':
            self.parameters.child('Signal').child('File').setValue(fname)

    def set_img_dir(self):
        """Opens file dialogue and sets the signal data folder
        """
        # ic()
        path = QFileDialog().getExistingDirectory(
            caption='Choose Image Directory',
            directory='',
            options=QFileDialog.ShowDirsOnly
        )
        if path != '':
            self.parameters.child('Signal').child('img_dir').setValue(path)
        self.img_dir = path

    def get_img_fname(self):
        """Sets file name based on chosen options
        """
        ic()
        old_fname = self.img_file
        if self.inp_type != 'Image Directory':
            img_file = self.parameters.child('Signal').child('File').value()
            if os.path.exists(img_file):
                self.img_file = img_file
                self.img_dir, _, self.img_ext = split_file_name(self.img_file)
                # self.meta_ext = self.get_meta_ext(self.img_file)

        else:
            self.img_ext = self.parameters.child('Signal').child('img_ext').value()
            self.img_dir = self.parameters.child('Signal').child('img_dir').value()
            self.include_subdir = self.parameters.child('Signal').child('include_subdir').value()

            filters = '*' + '*'.join(f for f in self.file_filter.split()) + '*'
            filters = filters if filters != '**' else '*'

            file_found = False
            ic(file_found, self.img_file)
            for idx, (subdir, dirs, files) in enumerate(os.walk(self.img_dir)):
                for file in files:
                    fname = os.path.join(subdir, file)
                    if fnmatch.fnmatch(fname, f'{filters}.{self.img_ext}'):
                        # ic(fname, self.img_file, self.poni_dict)
                        if match_img_detector(fname, self.poni_dict):
                            # ic(self.img_file, self.meta_ext)
                            if self.meta_ext:
                                if self.exists_meta_file(fname):
                                    self.img_file = fname
                                    file_found = True
                                    break
                                else:
                                    continue
                            else:
                                self.img_file = fname
                                break
                            # self.meta_ext = self.get_meta_ext(fname)
                ic(self.img_file, file_found, self.include_subdir, self.meta_ext)
                if file_found or (not self.include_subdir):
                    # ic('breaking')
                    break

        # ic(self.img_file, self.scan_parameters)
        if (((self.img_file != old_fname) or (self.img_file and (len(self.scan_parameters) < 1)))
                and self.meta_ext):
            ic('setting pars from meta')
            self.set_pars_from_meta()

    def set_series_average(self):
        self.series_average = self.parameters.child('Signal').child('series_average').value()

    def set_meta_ext(self):
        ic()
        self.meta_ext = self.parameters.child('Signal').child('meta_ext').value()
        if self.meta_ext == 'None':
            self.meta_ext = None
        self.get_img_fname()

    def exists_meta_file(self, img_file):
        """Checks for existence of meta file for image file"""
        ic(self.meta_ext)
        if self.meta_ext != 'SPEC':
            meta_files = [
                f'{os.path.splitext(img_file)[0]}.{self.meta_ext}',
                f'{img_file}.{self.meta_ext}'
            ]
            if os.path.exists(meta_files[0]) or os.path.exists(meta_files[1]):
                # ic('returning True \n')
                return True
        else:
            meta_file = get_specFile(img_file)
            if meta_file and os.path.exists(meta_file):
                return True

        return False

    def set_pars_from_meta(self):
        ic()
        self.get_scan_parameters()
        self.set_bg_matching_options()
        self.set_gi_motor_options()
        self.set_bg_norm_options()

    def set_mask_file(self):
        """Opens file dialogue and sets the mask file
        """
        fname, _ = QFileDialog().getOpenFileName(
            filter="EDF (*.edf)"
        )
        if fname != '':
            self.parameters.child('Signal').child('mask_file').setValue(fname)
        self.mask_file = fname

    def set_bg_type(self):
        """Change Parameter Names depending on BG Type
        """
        for child in self.parameters.child('BG').children():
            child.hide()
        self.parameters.child('BG').child('bg_type').show()

        self.bg_type = self.parameters.child('BG').child('bg_type').value()
        if self.bg_type == 'None':
            return
        elif self.bg_type != 'BG Directory':
            self.parameters.child('BG').child('File').show()
            if self.bg_type == 'Single BG File':
                opts = {'title': 'File Name'}
            else:
                opts = {'title': 'First File'}
            self.parameters.child('BG').child('File').setOpts(**opts)
            self.parameters.child('BG').child('bg_file_browse').show()
        else:
            self.parameters.child('BG').child('Match').show()

        self.parameters.child('BG').child('Scale').show()
        self.parameters.child('BG').child('norm_channel').show()

    def set_bg_file(self):
        """Opens file dialogue and sets the background file
        """
        # ic()
        fname, _ = QFileDialog().getOpenFileName(
            filter=f"Images (*.{self.img_ext})"
        )
        if fname != '':
            self.parameters.child('BG').child('File').setValue(fname)
        self.bg_file = fname

    def set_bg_dir(self):
        """Opens file dialogue and sets the background folder
        """
        # ic()
        path = QFileDialog().getExistingDirectory(
            caption='Choose BG Directory',
            directory='',
            options=QFileDialog.ShowDirsOnly
            # options =(QFileDialog.ShowDirsOnly)
        )
        if path != '':
            self.parameters.child('BG').child('Match').child('bg_dir').setValue(path)
        self.bg_dir = path

    def set_h5_dir(self):
        """Opens file dialogue and sets the path where processed data is stored
        """
        # ic()
        path = QFileDialog().getExistingDirectory(
            caption='Choose Save Directory',
            directory='',
            options=QFileDialog.ShowDirsOnly
        )
        if path != '':
            Path(path).mkdir(parents=True, exist_ok=True)
            self.parameters.child('h5_dir').setValue(path)
            self.h5_dir = path

    def set_bg_matching_options(self):
        """Reads image metadata to populate matching parameters
        """
        # ic(self.scan_parameters)
        if self.scan_parameters == []:
            return

        pars = [p for p in self.scan_parameters if not any(x.lower() in p.lower() for x in ['ROI', 'PD'])]
        pars.insert(0, 'None')
        if 'TEMP' in pars:
            pars.insert(1, pars.pop(pars.index('TEMP')))

        value = 'None'
        opts = {'values': pars, 'limits': pars, 'value': value}
        self.parameters.child('BG').child('Match').child('Parameter').setOpts(**opts)

    def set_bg_matching_par(self):
        """Changes bg matching parameter
        """
        # ic()
        self.bg_matching_par = self.parameters.child('BG').child('Match').child('Parameter').value()
        if self.bg_matching_par == 'None':
            self.bg_matching_par = None

    def set_bg_norm_options(self):
        """Counter Values used to normalize and subtract background
        """
        # ic()
        pars = self.counters
        # ic(self.counters)
        pars.insert(0, 'None')

        opts = {'values': pars, 'limits': pars, 'value': 'None'}
        self.parameters.child('BG').child('norm_channel').setOpts(**opts)

    def set_bg_norm_channel(self):
        """Changes bg matching parameter
        """
        # ic()
        self.bg_norm_channel = self.parameters.child('BG').child('norm_channel').value()

    def set_gi_motor_options(self):
        """Reads image metadata to populate possible GI theta motor
        """
        # ic()
        pars, value = ['Manual'], '0.1'
        pars += [p for p in self.motors if not any(x.lower() in p.lower() for x in ['ROI', 'PD'])]
        if 'th' in pars:
            pars.insert(0, pars.pop(pars.index('th')))
            value = 'th'
        elif 'theta' in pars:
            pars.insert(0, pars.pop(pars.index('theta')))
            value = 'theta'
        else:
            value = 'Theta'

        # pars = ['Manual'] + pars
        # ic(pars)

        opts = {'values': pars, 'limits': pars, 'value': value}
        self.parameters.child('GI').child('th_motor').setOpts(**opts)

    def set_gi_th_motor(self):
        """Update Grazing theta motor"""
        self.th_mtr = self.parameters.child('GI').child('th_motor').value()
        self.parameters.child('GI').child('th_val').hide()
        if self.th_mtr == 'Manual':
            self.parameters.child('GI').child('th_val').show()
            self.th_mtr = self.parameters.child('GI').child('th_val').value()
            # ic(self.th_mtr)

    def get_scan_parameters(self):
        """ Reads image metadata to populate matching parameters
        """
        ic(self.img_file, self.meta_ext)
        if not self.img_file:
            return

        img_meta = get_img_meta(self.img_file, self.meta_ext)
        self.scan_parameters = list(img_meta.keys())

        counters = get_img_meta(self.img_file, self.meta_ext, rv='Counters')
        self.counters = list(counters.keys())

        motors = get_img_meta(self.img_file, self.meta_ext, rv='Motors')
        self.motors = list(motors.keys())

    def enabled(self, enable):
        """Sets tree and start button to enable.

        args:
            enable: bool, True for enabled False for disabled.
        """
        # ic()
        self.tree.setEnabled(enable)
        self.ui.startButton.setEnabled(enable)

    def stylize_ParameterTree(self):
        self.tree.setStyleSheet("""
        QTreeView::item:has-children {
            background-color: rgb(230, 230, 230); 
            color: rgb(30, 30, 30);
        }
            """)


class specThread(wranglerThread):
    """Thread for controlling the specProcessor process. Receives
    manages a command and signal queue to pass commands from the main
    thread and communicate back relevant signals

    attributes:
        command_q: mp.Queue, queue to send commands to process
        file_lock: mp.Condition, process safe lock for file access
        scan_name: str, name of current scan
        fname: str, full path to data file.
        h5_dir: str, data file directory.
        img_file: str, path to image file
        img_dir: str, path to image directory
        img_ext : str, extension of image file
        series_average : bool, flag to average over series
        meta_ext : str, extension of metadata file
        poni_dict: str, Poni File name
        detector: str, Detector name
        input_q: mp.Queue, queue for commands sent from parent
        signal_q: mp.Queue, queue for commands sent from process
        sphere_args: dict, used as **kwargs in sphere initialization.
            see EwaldSphere.
        timeout: float or int, how long to continue checking for new
            data.
        command: command passed to stop, pause, continue etc.
        data_1d/2d: Dictionaries to store processed data for plotting

    signals:
        showLabel: str, sends out text to be used in specLabel

    methods:
        run: Main method, called by start
    """
    showLabel = Qt.QtCore.Signal(str)

    def __init__(
            self,
            command_queue,
            sphere_args,
            file_lock,
            fname,
            h5_dir,
            scan_name,
            single_img,
            poni_dict,
            inp_type,
            img_file,
            img_dir,
            include_subdir,
            img_ext,
            series_average,
            meta_ext,
            file_filter,
            mask_file,
            write_mode,
            bg_type,
            bg_file,
            bg_dir,
            bg_matching_par,
            bg_match_fname,
            bg_file_filter,
            bg_scale,
            bg_norm_channel,
            gi,
            th_mtr,
            timeout,
            command,
            sphere,
            data_1d,
            data_2d,
            parent=None):

        """command_queue: mp.Queue, queue for commands sent from parent
        sphere_args: dict, used as **kwargs in sphere initialization.
            see EwaldSphere.
        fname: str, path to data file.
        h5_dir: str, data file directory.
        file_lock: mp.Condition, process safe lock for file access
        scan_name: str, name of current scan
        single_img: bool, True if there is only one image
        poni_dict: str, poni file name
        detector: str, Detector name
        img_file: str, path to input image file
        img_dir: str, path to image directory
        include_subdir: bool, flag to include subdirectories
        img_ext : str, extension of image file
        series_average : bool, flag to average over series
        meta_ext : str, extension of metadata file
        timeout: float or int, how long to continue checking for new
            data.
        command: command passed to stop, pause, continue etc.
        gi: bool, grazing incidence flag to determine if pyGIX is to be used
        th_mtr: float, incidence angle
        """
        super().__init__(command_queue, sphere_args, fname, file_lock, parent)

        self.h5_dir = h5_dir
        self.scan_name = scan_name
        self.single_img = single_img
        self.poni_dict = poni_dict
        self.inp_type = inp_type
        self.img_file = img_file
        self.img_dir = img_dir
        self.include_subdir = include_subdir
        self.img_ext = img_ext
        self.series_average = series_average
        self.meta_ext = meta_ext
        self.file_filter = file_filter
        self.mask_file = mask_file
        self.write_mode = write_mode
        self.bg_type = bg_type
        self.bg_file = bg_file
        self.bg_dir = bg_dir
        self.bg_matching_par = bg_matching_par
        self.bg_match_fname = bg_match_fname
        self.bg_file_filter = bg_file_filter
        self.bg_scale = bg_scale
        self.bg_norm_channel = bg_norm_channel
        self.gi = gi
        self.th_mtr = th_mtr
        self.timeout = timeout
        self.command = command
        self.sphere = sphere
        self.data_1d = data_1d
        self.data_2d = data_2d

        self.user = None
        self.mask = None
        self.detector = None
        self.img_fnames = []
        self.processed = []
        self.processed_scans = []
        self.sub_label = ''

    def run(self):
        """Initializes specProcess and watches for new commands from
        parent or signals from the process.
        """
        t0 = time.time()
        if (self.poni_dict == '') or (self.img_file == ''):
            return

        self.img_fnames.clear()
        self.processed.clear()
        self.processed_scans.clear()
        self.detector = self.poni_dict['detector']
        self.sub_label = ''
        # self.get_mask()
        self.mask = get_mask_array(self.detector, self.mask_file)

        self.process_scan()
        print(f'Total Time: {time.time() - t0:0.2f}')

    def process_scan(self):
        """Go through series of images in a scan and process them individually
        """
        sphere = None
        files_processed = 0

        start = time.time()
        # start1 = time.time()
        # print(f'Start time: {time.time() - start1:0.2f}')
        while True:
            # Check for commands, or wait if paused
            command = self.command
            # ic(command)
            if command == 'stop':
                break
            elif command == 'pause':
                self.showLabel.emit(f'Paused')
                time.sleep(0.5)
                continue
            else:
                pass

            # img_file, img_number, img_data = self.get_next_image()
            img_file, scan_name, img_number, img_data, img_meta = self.get_next_image()
            if img_data is None:
                if img_file is None:
                    self.showLabel.emit(f'Checking for next image')
                    time.sleep(0.2)
                else:
                    print(f'Invalid Image File {os.path.basename(img_file)}. Skipping...')

                elapsed = time.time() - start
                if elapsed > self.timeout:
                    self.showLabel.emit(f'Timeout occurred')
                    break
                else:
                    continue
            else:
                fname = os.path.splitext(os.path.basename(img_file))[0]
                self.showLabel.emit(f'Processing {fname[-30:]}')

            # self.scan_name = get_scan_name(img_file)
            img_number = 1 if (img_number is None) else img_number
            self.scan_name = scan_name

            # Initialize sphere and save to disk, send update for new scan
            if (sphere is None) or (self.scan_name != sphere.name):
                sphere = self.initialize_sphere()

            if img_number in list(sphere.arches.index):
                if self.single_img:
                    self.sigUpdate.emit(img_number)
                    break
                continue

            # Get Background
            bg_raw = self.get_background(img_file, img_number, img_meta)

            # ic(self.th_mtr)

            # Initialize arch and integrate
            arch = EwaldArch(
                img_number, img_data, poni_dict=self.poni_dict,
                scan_info=img_meta, static=True, gi=self.gi,
                th_mtr=self.th_mtr, bg_raw=bg_raw,
                series_average=self.series_average
            )

            # integrate image to 1d and 2d arrays
            arch.integrate_1d(global_mask=self.mask, **sphere.bai_1d_args)
            arch.integrate_2d(global_mask=self.mask, **sphere.bai_2d_args)

            # Add arch copy to sphere, save to file
            with self.file_lock:
                sphere.add_arch(
                    arch=arch, calculate=False, update=True,
                    get_sd=True, set_mg=False, static=True, gi=self.gi,
                    th_mtr=self.th_mtr, series_average=self.series_average
                )
                sphere.save_to_h5(data_only=True, replace=False)
            # print(f'Saved data to h5: {time.time() - start1:0.2f}')

            # Save 1D integrated data in CSV and xye files
            self.save_1d(sphere, arch, img_number)
            # print(f'Saved 1D data: {time.time() - start1:0.2f}')

            print(f'Processed {fname} {self.sub_label}')
            if len(fname) > 40:
                fname = f'{fname[:8]}....{fname[-30:]}'
            self.showLabel.emit(f'{fname}')
            self.sigUpdate.emit(img_number)
            files_processed += 1

            if self.single_img:
                self.sigUpdate.emit(img_number)
                break

            time.sleep(0.02)
            start = time.time()

        # If loop ends, signal terminate to parent thread.
        print(f'\nTotal Files Processed: {files_processed}')

    def get_next_image(self):
        """ Gets next image in image series or in directory to process

        Returns:
            image_name {str}: image file path
            image_number {int}: image file number (if part of series)
            image_data {np.ndarray}: image file data array
        """
        if self.single_img:
            img_data = get_img_data(self.img_file, self.detector, return_float=True)
            meta = get_img_meta(self.img_file, self.meta_ext) if self.meta_ext else {}
            # return self.img_file, get_img_number(self.img_file), img_data
            scan_name, img_number = get_sname_img_number(self.img_file)
            return self.img_file, scan_name, img_number, img_data, meta

        if len(self.img_fnames) == 0:
            if self.inp_type != 'Image Directory':
                first_img = self.img_file
                self.img_fnames = Path(self.img_dir).glob(f'{self.scan_name}_*.{self.img_ext}')
            else:
                first_img = ''
                filters = '*' + '*'.join(f for f in self.file_filter.split()) + '*'
                filters = filters if filters != '**' else '*'
                if self.include_subdir:
                    self.img_fnames = Path(self.img_dir).rglob(f'{filters}.{self.img_ext}')
                else:
                    self.img_fnames = Path(self.img_dir).glob(f'{filters}.{self.img_ext}')

            self.img_fnames = [str(f) for f in self.img_fnames if
                               (str(f) >= first_img) and (str(f) not in self.processed)]

            # self.img_fnames = deque(sorted(self.img_fnames))
            self.img_fnames = deque(natural_sort_ints(self.img_fnames))

        img_file, scan_name, img_number, img_data, img_meta = None, None, 1, None, {}
        n = 0
        while len(self.img_fnames) > 0:
            fname = self.img_fnames[0]
            sname, snumber = get_sname_img_number(fname)

            if (n > 0) and (scan_name != sname):
                break

            self.processed.append(fname)
            self.img_fnames.popleft()

            data = get_img_data(fname, self.detector, return_float=True)
            if data is None:
                continue

            meta = get_img_meta(fname, self.meta_ext) if self.meta_ext else {}
            n += 1

            if (not self.series_average) or (snumber is None):
                return fname, sname, snumber, data, meta
            else:
                if n == 1:
                    img_data = data
                    img_meta = meta
                else:
                    # img_data = img_data*(n-1)/n + data/n
                    img_data += data
                    for (k, v) in meta.items():
                        try:
                            img_meta[k] = float(img_meta[k]) + float(meta[k])
                        except TypeError:
                            pass

                scan_name, img_file = sname, fname
                # ic(sname, scan_name, n)
                # if len(self.img_fnames) == 0:
                #     return img_file, scan_name, 1, img_data, meta

        if n > 1:
            img_data /= n
            for (k, v) in img_meta.items():
                try:
                    img_meta[k] /= n
                except TypeError:
                    pass

        return img_file, scan_name, img_number, img_data, img_meta
        # return None, None, None, None, None

    def get_meta_data(self, img_file):
        meta_file = f'{os.path.splitext(img_file)[0]}.{self.meta_ext}'
        return get_img_meta(meta_file)

    def subtract_bg(self, img_data, img_file, img_number, img_meta):
        bg = self.get_background(img_file, img_number, img_meta)
        try:
            img_data -= bg
            # min_int = img_data.min()
            # if min_int < 0:
            #     img_data -= min_int
        except ValueError:
            pass

    def initialize_sphere(self):
        """ If scan changes, initialize new EwaldSphere object
        If mode is overwrite, replace existing HDF5 file, else append to it
        """
        fname = os.path.join(self.h5_dir, self.scan_name + '.hdf5')
        sphere = EwaldSphere(self.scan_name,
                             data_file=fname,
                             static=True,
                             gi=self.gi,
                             th_mtr=self.th_mtr,
                             series_average=self.series_average,
                             single_img=self.single_img,
                             global_mask=self.mask,
                             **self.sphere_args)

        write_mode = self.write_mode
        if not os.path.exists(fname):
            write_mode = 'Overwrite'

        with self.file_lock:
            if write_mode == 'Append':
                sphere.load_from_h5(replace=False, mode='a')
                for (k, v) in self.sphere_args.items():
                    setattr(sphere, k, v)
                existing_arches = sphere.arches.index
                if len(existing_arches) == 0:
                    sphere.save_to_h5(replace=True)
            else:
                sphere.save_to_h5(replace=True)

        self.sigUpdateFile.emit(
            self.scan_name, fname,
            self.gi, self.th_mtr, self.single_img,
            self.series_average
        )
        print(f'\n***** New Scan *****')

        return sphere

    def get_mask(self):
        """Get mask array from mask file
        """
        self.mask = self.detector.calc_mask()
        if self.mask_file and os.path.exists(self.mask_file):
            if self.mask is not None:
                try:
                    self.mask += fabio.open(self.mask_file).data
                except ValueError:
                    print('Mask file not valid for Detector')
                    pass
            else:
                self.mask = fabio.open(self.mask_file).data

        if self.mask is None:
            return None

        if self.mask.shape != self.detector.shape:
            print('Mask file not valid for Detector')
            return None

        self.mask = np.flatnonzero(self.mask)

    def get_background(self, img_file, img_number, img_meta):
        """Subtract background image if bg_file or bg_dir specified
        """
        if self.bg_type == 'None':
            return 0

        bg, bg_file, bg_meta, norm_factor = 0, None, None, 1
        self.sub_label, norm_label, bg_scale_label = '', '', ''

        if self.bg_type == 'Single BG File':
            if self.bg_file:
                bg_file = self.bg_file
                bg_meta = get_img_meta(bg_file, self.meta_ext)
        elif self.bg_type == 'Series Average':
            if self.bg_file:
                sname, fnames, bg, bg_meta = get_series_avg(self.bg_file, self.detector, self.meta_ext)
                if sname is None:
                    return 0
        else:
            if self.bg_dir and (self.bg_match_fname or self.bg_matching_par):
                bg_file_filter = 'bg' if not self.bg_file_filter else self.bg_file_filter
                if self.bg_match_fname:
                    bg_file_filter = f'{self.scan_name} {bg_file_filter}'
                filters = '*' + '*'.join(f for f in bg_file_filter.split()) + '*'
                filters = filters if filters != '**' else '*'

                meta_files = sorted(glob.glob(os.path.join(
                    # self.img_dir, f'{filters}[0-9][0-9][0-9][0-9].{self.meta_ext}')))
                    self.img_dir, f'{filters}.{self.meta_ext}')))

                for meta_file in meta_files:
                    bg_file = f'{os.path.splitext(meta_file)[0]}.{self.img_ext}'
                    if bg_file == img_file:
                        bg_file = None
                        continue

                    # bg_meta = get_img_meta(meta_file)
                    bg_meta = get_img_meta(bg_file, self.meta_ext)
                    if self.bg_match_fname:
                        if img_number == get_img_number(meta_file):
                            break
                    else:
                        try:
                            if bg_meta[self.bg_matching_par] == img_meta[self.bg_matching_par]:
                                break
                        except KeyError:
                            bg_file = None
                            continue

        if self.bg_type != 'Series Average':
            if bg_file is None:
                return 0.

            bg = get_img_data(bg_file, self.detector, return_float=True)
            if bg is None:
                return 0.

        if self.bg_scale != 1:
            bg *= self.bg_scale
            bg_scale_label = f'{self.bg_scale:0.2f} [Scale] x '
        if (self.bg_norm_channel != 'None') and (img_meta is not None) and (bg_meta is not None):
            try:
                if ((self.bg_norm_channel in img_meta.keys()) and
                        (self.bg_norm_channel in bg_meta.keys()) and
                        (bg_meta[self.bg_norm_channel] != 0)):
                    norm_factor = (img_meta[self.bg_norm_channel] / bg_meta[self.bg_norm_channel])
                    bg *= norm_factor
                    norm_label = f'{norm_factor:0.2f} [Normalized to Channel - {self.bg_norm_channel}] x '
            except (KeyError, TypeError):
                pass

        if self.bg_type != 'Series Average':
            self.sub_label = f'[Subtracted {bg_scale_label}{norm_label}{os.path.basename(bg_file)}]'
        else:
            self.sub_label = f'[Subtracted {bg_scale_label}{norm_label}{sname}]'

        return bg

    @staticmethod
    def save_1d(sphere, arch, idx):
        """
        Automatically save 1D integrated data
        """
        path = os.path.dirname(sphere.data_file)
        path = os.path.join(path, sphere.name)
        Path(path).mkdir(parents=True, exist_ok=True)

        q, tth, intensity = arch.int_1d.q, arch.int_1d.ttheta, arch.int_1d.norm

        # Write I(q) to xye
        fname = os.path.join(path, f'iq_{sphere.name}_{str(idx).zfill(4)}.xye')
        write_xye(fname, q, intensity, np.sqrt(abs(intensity)))

        # Write I(tth) to xye
        fname = os.path.join(path, f'itth_{sphere.name}_{str(idx).zfill(4)}.xye')
        write_xye(fname, tth, intensity, np.sqrt(abs(intensity)))

        # Write I(q) to csv
        fname = os.path.join(path, f'iq_{sphere.name}_{str(idx).zfill(4)}.csv')
        write_csv(fname, q, intensity)

        # Write I(tth) to csv
        fname = os.path.join(path, f'itth_{sphere.name}_{str(idx).zfill(4)}.csv')
        write_csv(fname, tth, intensity)


def atoi(text):
    return int(text) if text.isdigit() else text


def natural_keys_int(text):
    """
    alist.sort(key=natural_keys) sorts in human order
    http://nedbatchelder.com/blog/200712/human_sorting.html
    (See Toothy's implementation in the comments)
    """
    return [atoi(c) for c in re.split(r'(\d+)', text)]


def atof(text):
    try:
        retval = float(text)
    except ValueError:
        retval = text
    return retval


def natural_keys_float(text):
    """
    alist.sort(key=natural_keys) sorts in human order
    http://nedbatchelder.com/blog/200712/human_sorting.html
    (See Toothy's implementation in the comments)
    float regex comes from https://stackoverflow.com/a/12643073/190597
    """
    return [atof(c) for c in re.split(r'[+-]?([0-9]+(?:[.][0-9]*)?|[.][0-9]+)', text)]


def natural_sort_ints(list_to_sort):
    return sorted(list_to_sort, key=natural_keys_int)


def natural_sort_float(list_to_sort):
    return sorted(list_to_sort, key=natural_keys_float)




import asyncio
import sys, traceback


class ControlFrame(FrameObject):

    def __init__(self, frame_pool, mi_list, var_list=None):
        """run the common init routine to create and register futures, generate the UUID"""
        self._common_init(frame_pool, mi_list)

        #Init. the list of followers as empty.
        self.updated_variables = []

        #Finally, set the allow_attachement
        #self.lock_and_set_attachable(True)
        self.allow_attachement = True


        #Register myself
        self.attach_pulpits()

    def set_variables_to_update(self, var_to_follow):
        """Defines the variables that are updated by this control frame"""
        self.updated_variables = var_to_follow

    def allows_followers(self, required_var_list):
        """Check if this frame can follow all of the required_vars. Control frames start with an empty list that is
        updated by the async routine actually manipulating the motors calling the set_variables_to_update"""

        try:
            for var in required_var_list:
                #Check that the current variable has a motor interactor in the oracle frames list of motor interactors
                self.updated_variables.index(var)
        except ValueError:
            #This means the variable is not in the control frames update list
            return False
        else:
            #All of the required_vars are within the scope of this control frame
            return True


    def start_move_position(self, move_to):
        """Launches the asyncronous routine that actually moves the motor position. """
        #Start the oracle frame async routine.
        self.running_future = asyncio.ensure_future(self.run_move_position(move_to), loop=self.frame_pool.loop)

        for mi in self.mi_list:
            self.updated_variables.append(mi.motor_variables.position)
            self.updated_variables.append(mi.motor_variables.motor_moving)
            self.updated_variables.append(mi.motor_variables.error_flag)
            self.updated_variables.append(mi.motor_variables.error_register)
            self.updated_variables.append(mi.motor_variables.hMT_leadlag_count)
            self.updated_variables.append(mi.motor_variables.hMT_actual_velocity)
        return self.running_future


    @FrameObject.add_pulpit_control_flow
    @FrameObject.handle_communication_errors
    @FrameObject.update_orphaned_followers_afterwards
    async def run_move_position(self, move_to): #Need to add variables to update
        """Run the absolute move routine for a single motor interactor. This routine will check the motor_moving flag util it is set back to zero. This blocks pulpit until complete"""
        debug('run_move_position: starting')
        t0 = time.time()
        mi = self.mi_list[0]

        status = await mi.check_status_and_errors(self.frame_uuid)

        print("Fast_hMT_status: ", status)

        #Execute the move command for this motor
        print("Moving motor interactor '{}' to '{}'".format(mi, move_to))
        await mi.absolute_move(move_to)

        #Poll the motor to find when it stops moving
        while True:
            status = await mi.check_status_and_errors(self.frame_uuid)
            print("Motor status: ", status)

            if status['motor_moving'] is False:
                break

        print("   done in {}s".format(time.time() - t0))


    @FrameObject.add_pulpit_control_flow
    @FrameObject.repeat_until_complete
    @FrameObject.handle_communication_errors
    @FrameObject.handle_motor_config_error
    @FrameObject.update_orphaned_followers_afterwards
    async def run_move_to_indexed_position(self, move_to_index):
        """This will move to an indexed position for my kinematic switching setup"""
        debug('run_move_to_indexed_position: starting')
        t0=time.time()

        cam_mi = self.mi_list[0]
        disc_mi = self.mi_list[1]

        #Do a read on each motor to make sure it is still alive - may not be necessary.
        print("calling fast_hMT status on each motor to check if alive")
        for mi in self.mi_list:
            status = await mi.check_status_and_errors(self.frame_uuid)

        print(" Elapsed time {} seconds".format(time.time()-t0))

        #Start by moving the cam out
        await cam_mi.absolute_move_and_wait(12800, frame_uuid=self.frame_uuid, also_fast_update=[disc_mi])
        print(" Elapsed time {} seconds".format(time.time() - t0))

        #Rotate the disc to the correct index
        await disc_mi.absolute_move_and_wait(25600 * move_to_index, frame_uuid=self.frame_uuid, also_fast_update=[cam_mi])
        print(" Elapsed time {} seconds".format(time.time() - t0))

        # Move the cam back to the locked position
        await cam_mi.absolute_move_and_wait(0, frame_uuid=self.frame_uuid, also_fast_update=[disc_mi])
        print(" Elapsed time {} seconds".format(time.time() - t0))


    @FrameObject.add_pulpit_control_flow
    @FrameObject.repeat_until_complete
    @FrameObject.handle_communication_errors
    @FrameObject.handle_motor_config_error
    @FrameObject.update_orphaned_followers_afterwards
    async def run_get_indexed_position(self):
        """This is a control frame that returns the indexed position of the kinematic switching setup. This is a
        control frame because we want exclusive control of the motor interactors to make sure things are not actually
        moving during the measurement"""
        debug('run_get_index_position: starting')
        t0=time.time()

        cam_mi = self.mi_list[0]
        disc_mi = self.mi_list[1]

        #Do a read to get the motor positions
        cam_status = await cam_mi.check_status_and_errors(self.frame_uuid)
        disc_status = await disc_mi.check_status_and_errors(self.frame_uuid)

        #Position right now is just a simple function
        self.return_result = int( disc_status['position'] / 25600)




import traceback
from multiprocessing.managers import SharedMemoryManager
from multiprocessing.shared_memory import ShareableList, SharedMemory
from multiprocessing import RLock

import numpy as np

from .smbase import SMBase


def shape_to_bytes(shape):
    out = list(shape)
    out.append(-1)
    return np.array(out, dtype=int).tobytes()


def bytes_to_shape(buf):
    out = np.frombuffer(buf, dtype=int)
    return tuple(out[:-1])


class SMArray(SMBase):
    def __init__(self, array=None, addr=None, mutex=None, shape=(0,), dtype='int32',
                 manager=None, ratio=1):
        itemsize = np.dtype(dtype).itemsize
        size = int(itemsize * np.prod(shape))
        if size < itemsize * 2:
            size = itemsize * 2
        shape_length = len(shape)
        if shape_length < 16:
            shape_length = 16
        format_list = [
            '0' * 128,
            size,
            size,
            shape_to_bytes(list(range(shape_length))),
            '000'
        ]
        super().__init__(addr, manager, mutex, format_list, size, ratio)
        with self.mutex:
            if addr is None:
                self._shl[3] = shape_to_bytes(shape)
                self._shl[4] = np.dtype(dtype).char
            self._dtype = np.dtype(self._shl[4])
            self.npview = np.ndarray(
                bytes_to_shape(self._shl[3]),
                dtype=self._dtype,
                buffer=self._shm.buf
            )

    def check_memory(self) -> bool:
        with self.mutex:
            updated = super().check_memory()
            if self._dtype != np.dtype(self._shl[4]):
                self._dtype = np.dtype(self._shl[4])
                updated = True
            if self.npview.shape != bytes_to_shape(self._shl[3]):
                updated = True
            if updated:
                self.npview = np.ndarray(
                    bytes_to_shape(self._shl[3]),
                    dtype=self._dtype,
                    buffer=self._shm.buf
                )
        return updated

    def _recap(self, cap):
        with self.mutex:
            super()._recap(cap)
            if cap > self._shm.size:
                self.npview = np.ndarray(
                    bytes_to_shape(self._shl[3]),
                    dtype=self._dtype,
                    buffer=self._shm.buf
                )

    def set_dtype(self, dtype, resize=False):
        if np.dtype(dtype) != self._dtype:
            with self.mutex:
                self.check_memory()
                needed_cap = np.dtype(dtype).itemsize * np.prod(self.npview.shape)
                if needed_cap > self._shm.size:
                    if resize:
                        self._recap(needed_cap)
                    else:
                        raise ValueError("New dtype overflows buffer")
                self.npview = np.ndarray(self.npview.shape, dtype=dtype,
                                         buffer=self._shm.buf)
                self._dtype = np.dtype(dtype)
                self._shl[4] = self._dtype.char

    def reshape(self, shape, resize=False):
        with self.mutex:
            self.check_memory()
            if np.prod(shape) != np.prod(self.npview.shape):
                if not resize:
                    raise ValueError("New shape would change size, but resize flag is False")
                needed_cap = self.npview.dtype.itemsize * np.prod(shape)
                self._recap(needed_cap)
                self.npview = np.ndarray(shape, dtype=self._dtype,
                                         buffer=self._shm.buf)
            else:
                self.npview = self.npview.reshape(shape)
            self._shl[3] = shape_to_bytes(shape)


class SMArrayDescriptor:
    def __init__(self, signal=None):
        if signal is None:
            self.signal_emit = lambda x: None
        else:
            self.signal_emit = signal

    def __set_name__(self, owner, name):
        self.private_name = '_' + name

    def __get__(self, instance, owner):
        sma: SMArray = getattr(instance, self.private_name)
        update = sma.check_memory()
        if update:
            self.signal_emit(self.private_name)
        return sma.npview

    def __set__(self, instance, value):
        if isinstance(value, SMArray):
            setattr(instance, self.private_name, value)
        elif isinstance(value, np.ndarray):
            sma: SMArray = getattr(instance, self.private_name)
            sma.check_memory()
            if sma.npview.dtype != value.dtype:
                sma.set_dtype(value.dtype, resize=True)
            if sma.npview.shape != value.shape:
                sma.reshape(value.shape, resize=True)
            sma.npview[:] = value[:]
        else:
            raise TypeError(f"Attribute {self.private_name} must be set by " +
                            "ndarray or SMArray")





from collections import namedtuple
from dataclasses import dataclass, field
import copy

import numpy as np
from pyFAI import units
import h5py

from .. import _utils as utils


class nzarray1d():
    """Sparse matrix like object which stores minimal box to contain
    non-zero data. Only for 1D arrays.
    
    attributes:
        corners: tuple, edges of the region containing non-zero data
        data: numpy array, region with non-zero data. May also contain zeros.
        shape: tuple, shape of the full dataset
    
    methods:
        Reimplements all arithmetic functions (+, -, *, /, //)
        from_hdf5: Loads data from an hdf5 file
        full: returns the full dataset
        get_corners: Finds the edges of the non-zero region
        get_data: Cuts out the non-zero region
        get_shape: Gets the shape of the full dataset
        intersect: Finds the interesection between two nzarray1d
            objects
        to_hdf5: Saves the data to an hdf5 file
    """
    def __init__(self, arr=None, grp=None, lazy=False):
        """arr: numpy array, full dataset
        grp: h5py File or Group object, if used will load in data
        lazy: bool, if True the data attribute is a view of the
            h5py dataset called 'data' in grp
        """
        self.none_flag = False
        if isinstance(arr, self.__class__):
            if arr.data is None:
                self._none_array()
                self.none_flag = True
            else:
                self.data = np.empty_like(arr.data)
                self.data[()] = arr.data[()]
                self.shape = copy.deepcopy(arr.shape)
                self.corners = copy.deepcopy(arr.corners)
        elif grp is not None:
            if lazy:
                self.shape = grp['shape'][()]
                self.corners = grp['corners'][()]
                self.data = grp['data']
            else:
                self.shape = grp['shape'][()]
                self.corners = grp['corners'][()]
                self.data = grp['data'][()]
        elif arr is None:
            self.none_flag = True
            self._none_array()
        else:
            self.shape = self.get_shape(arr)
            self.corners = self.get_corners(arr)
            self.data = self.get_data(arr)
    
    def _none_array(self):
        arrn = np.array([0])
        self.shape = self.get_shape(arrn)
        self.corners = self.get_corners(arrn)
        self.data = self.get_data(arrn)
    
    def get_shape(self, arr):
        """Finds the shape of arr, enforces 1D arrays only. 
        
        args:
            arr: numpy array
        
        returns:
            shape: tuple, shape of the array
        """
        assert len(arr.shape) == 1, "Must be 1d array."
        return arr.shape
    
    def get_corners(self, arr):
        """Finds edges of non-zero region.
        
        args:
            arr: numpy array
        
        returns:
            corners: tuple, edges of the non-zero region
        """
        if (arr == 0).all():
            return [0,0]
        else:
            c = np.nonzero(arr)[0]
            return (c[0], c[-1] + 1)
    
    def get_data(self, arr):
        """Gets non-zero region of dataset.
        
        args:
            arr: numpy array
        
        returns:
            data: numpy array, non-zero region
        """
        data = arr[
            self.corners[0]:self.corners[1]
        ]
        return data
    
    def full(self):
        """Returns the full dataset.
        
        returns:
            full: numpy array, dataset with shape self.shape
        """
        if self.data is None:
            return None
        full = np.zeros(self.shape, dtype=self.data.dtype)
        full[self.corners[0]:self.corners[1]] = self.data
        return full

    def intersect(self, other):
        """Finds the intersection between two nzarrays. Returns a
        nzarray1d object with data region of a minimal size to include
        the full non-zero region from self and other.
        
        args:
            other: nzarray1d, object to compare against
        
        returns:
            out: nzarray1d, interection array
        """
        assert \
            self.shape == other.shape, \
            "Can't cast nzarray of different shape"
        out = nzarray1d()
        
        out.shape = self.shape[:]
        out.corners = [min(self.corners[0], other.corners[0]),
                    max(self.corners[1], other.corners[1])]
        out.data = np.zeros((out.corners[1] - out.corners[0],))

        other_data = np.zeros_like(out.data)
        
        i0, i1 = out._shift_index(self.corners)
        out.data[i0:i1] = self.data
        
        i0, i1 = out._shift_index(other.corners)
        other_data[i0:i1] = other.data
        
        return out, other_data
    
    def to_hdf5(self, grp, compression=None):
        """Saves data to an hdf5 file.
        
        args:
            grp: h5py File or Group, where data will be saved
            compression: str, compression algorithm to use. See h5py
                docs.
        """
        grp.attrs['encoded'] = 'nzarray'
        for name in ['shape', 'corners']:
            data = getattr(self, name)
            if name in grp:
                grp[name][()] = np.array(data)[()]
            else:
                grp.create_dataset(name, data=data)
        if 'data' in grp:
            grp['data'].resize(self.data.shape)
            grp['data'][()] = self.data[()]
        else:
            grp.create_dataset(
                'data', data=self.data, compression=compression, chunks=True,
                maxshape=tuple(
                    None for x in self.data.shape
                ), dtype='float64'
            )
    
    def from_hdf5(self, grp):
        """Loads in data from hdf5 file.
        
        args:
            grp: h5py File or Group, where to load data from.
        """
        for key in ['data', 'shape', 'corners']:
            self.__dict__[key] = np.empty_like(grp[key])
            self.__dict__[key] = grp[key][()]
        
    
    def _shift_index(self, idx, si=0):
        """Shifts index values for full dataset to query data from
        self.
        
        args:
            idx: iterable, indexes to shift
            si: int, starting index
        
        returns:
            out: list, new set of shifted indeces
        """
        out = []
        for i, val in enumerate(idx):
            j = i - i % 2 + si
            idx = self._get_idx(val, j)
            out.append(idx - self.corners[j])
        return out
    
    def _get_idx(self, x, i=0):
        """Handles negative indexing.
        
        args:
            x: int, index value
            i: int, which value in shape to use to handle negative
                indexes
        """
        if x >= 0:
            idx = x
        else:
            idx = self.shape[i] + x
        return idx
    
    def _shift_slice(self, key, i=0):
        """Shifts a slice object to allow data access without needing
        to call the full dataset.
        
        args:
            key: slice, desired slice of data
            i: int, dimension in shape to use
        
        returns:
            slice, bool: Either the original slice or shifted slice,
                depending on whether the slice is entirely contained in
                the dataset or not.
        """
        if key.start is None or key.stop is None:
            return key, True
        elif (self._get_idx(key.start, i) >= self.corners[0] and 
              self._get_idx(key.stop, i) <= self.corners[1]):
            start, stop = self._shift_index([key.start, key.stop], i)
            return slice(start, stop, key.step), False
        else:
            return key, True
    
    def __getitem__(self, key):
        if isinstance(key, slice):
            skey, full = self._shift_slice(key)
            if full:
                return self.full()[skey]
            else:
                return self.data[skey]
        elif type(key) == int:
            idx = self._get_idx(key)
            if self.corners[0] <= idx < self.corners[1]:
                return self.data[idx - self.corners[0]]
            else:
                return 0
        else:
            return self.full()[key]
    
    def __set__(self, obj, arr):
        if isinstance(arr, self.__class__):
            self.data = arr.data
            self.shape = arr.shape
            self.corners = arr.corners
        elif arr is None:
            self._none_array()
        else:
            self.shape = self.get_shape(arr)
            self.corners = self.get_corners(arr)
            self.data = self.get_data(arr)
    
    def __add__(self, other):
        if isinstance(other, self.__class__):
            assert list(self.shape) == list(other.shape), "Cannot add arrays of different shape"
            if other.data.size > 0 and self.data.size > 0:
                out, temp = self.intersect(other)
                out.data += temp
            elif other.data.size > 0:
                out = self.__class__(other)
            else:
                out = self.__class__(self)
        elif np.isscalar(other) or type(other) == np.ndarray:
            out = self.__class__(self.full() + other)
        else:
            raise TypeError(f"Cannot add object of type {type(other)}")
        return out
    
    def __sub__(self, other):
        if isinstance(other, self.__class__):
            assert list(self.shape) == list(other.shape), "Cannot subtract arrays of different shape"
            if other.data.size > 0 and self.data.size > 0:
                out, temp = self.intersect(other)
                out.data -= temp
            elif other.data.size > 0:
                out = self.__class__(other * -1)
            else:
                out = self.__class__(self)
        elif np.isscalar(other) or type(other) == np.ndarray:
            out = self.__class__(self.full() - other)
        return out
    
    def __mul__(self, other):
        if isinstance(other, self.__class__):
            assert list(self.shape) == list(other.shape), "Cannot multiply arrays of different shape"
            out, temp = self.intersect(other)
            out.data *= temp
            if other.data.size > 0 and self.data.size > 0:
                out, temp = self.intersect(other)
                out.data *= temp
            elif other.data.size > 0:
                out = self.__class__(self)
            else:
                out = self.__class__(other)
        elif np.isscalar(other):
            if other == 0:
                out = self.__class__(np.zeros(self.shape))
            else:
                out = self.__class__(self)
                out.data *= other
        elif type(other) == np.ndarray:
            out = self.__class__(self.full() * other)
        return out
    
    def __div__(self, other):
        return self.__truediv__(other)
    
    def __truediv__(self, other):
        if isinstance(other, self.__class__):
            assert list(self.shape) == list(other.shape), "Cannot divide arrays of different shape"
            out, temp = self.intersect(other)
            out.data *= temp
            if other.data.size > 0 and self.data.size > 0:
                out, temp = self.intersect(other)
                out.data = utils.div0(out.data, temp)
            elif other.data.size > 0:
                out = self.__class__(self)
            else:
                out = self.__class__(other)
        elif np.isscalar(other):
            if other == 0:
                out = self.__class__(np.zeros(self.shape))
            else:
                out = self.__class__(self)
                out.data = utils.div0(out.data, other)
        elif type(other) == np.ndarray:
            out = self.__class__(utils.div0(self.full(), other))
        return out
    
    def __floordiv__(self, other):
        if isinstance(other, self.__class__):
            assert list(self.shape) == list(other.shape), "Cannot divide arrays of different shape"
            out, temp = self.intersect(other)
            out.data *= temp
            if other.data.size > 0 and self.data.size > 0:
                out, temp = self.intersect(other)
                out.data = utils.div0(out.data, temp).astype(int)
            elif other.data.size > 0:
                out = self.__class__(self)
            else:
                out = self.__class__(other)
        elif np.isscalar(other):
            if other == 0:
                out = self.__class__(np.zeros(self.shape))
            else:
                out = self.__class__(self)
                out.data = utils.div0(out.data, other).astype(int)
        elif type(other) == np.ndarray:
            out = self.__class__(utils.div0(self.full(), other).astype(int))
        return out


class nzarray2d(nzarray1d):
    """Sparse matrix like object which stores minimal box to contain
    non-zero data. Only for 2D arrays.
    
    attributes:
        corners: tuple, edges of the region containing non-zero data
        data: numpy array, region with non-zero data. May also contain zeros.
        shape: tuple, shape of the full dataset
    
    methods:
        Reimplements all arithmetic functions (+, -, *, /, //)
        from_hdf5: Loads data from an hdf5 file
        full: returns the full dataset
        get_corners: Finds the edges of the non-zero region
        get_data: Cuts out the non-zero region
        get_shape: Gets the shape of the full dataset
        intersect: Finds the interesection between two nzarray1d
            objects
        to_hdf5: Saves the data to an hdf5 file
    """
    def __init__(self, arr=None, grp=None, lazy=False):
        """arr: numpy array, full dataset
        grp: h5py File or Group object, if used will load in data
        lazy: bool, if True the data attribute is a view of the
            h5py dataset called 'data' in grp
        """
        super().__init__(arr, grp, lazy)
    
    def _none_array(self):
        arrn = np.array([[0],[0]])
        self.shape = self.get_shape(arrn)
        self.corners = self.get_corners(arrn)
        self.data = self.get_data(arrn)
    
    def get_shape(self, arr):
        """Finds the shape of arr, enforces 2D arrays only. 
        
        args:
            arr: numpy array
        
        returns:
            shape: tuple, shape of the array
        """
        assert len(arr.shape) == 2, 'Must be 2D array.'
        return arr.shape[:]
    
    def get_corners(self, arr):
        """Finds edges of non-zero region.
        
        args:
            arr: numpy array
        
        returns:
            corners: tuple, edges of the non-zero region
        """
        if (arr == 0).all():
            return [0,0,0,0]
        else:
            r = np.nonzero(arr)[0]
            c = np.nonzero(arr)[1]
            return (min(r), max(r) + 1, min(c), max(c) + 1)
    
    def get_data(self, arr):
        """Gets non-zero region of dataset.
        
        args:
            arr: numpy array
        
        returns:
            data: numpy array, non-zero region
        """
        data = arr[
            self.corners[0]:self.corners[1], 
            self.corners[2]:self.corners[3]
        ]
        return data
    
    def full(self):
        """Returns the full dataset.
        
        returns:
            full: numpy array, dataset with shape self.shape
        """
        if self.data is None:
            return None
        arr = np.zeros(self.shape, dtype=self.data.dtype)
        arr[
            self.corners[0]:self.corners[1], 
            self.corners[2]:self.corners[3]
        ] = self.data
        return arr
    
    def intersect(self, other):
        """Finds the intersection between two nzarrays. Returns a
        nzarray2d object with data region of a minimal size to include
        the full non-zero region from self and other.
        
        args:
            other: nzarray2d, object to compare against
        
        returns:
            out: nzarray2d, interection array
        """
        assert list(self.shape) == list(other.shape), "Can't cast nzarray of different shape"
        out = nzarray2d()
        
        out.shape = self.shape[:]
        out.corners = [min(self.corners[0], other.corners[0]),
                       max(self.corners[1], other.corners[1]),
                       min(self.corners[2], other.corners[2]),
                       max(self.corners[3], other.corners[3])]
        
        out.data = np.zeros((out.corners[1] - out.corners[0],
                         out.corners[3] - out.corners[2]))
        
        other_data = np.zeros_like(out.data)
        
        r0, r1, c0, c1 = out._shift_index(self.corners)
        
        out.data[r0:r1,c0:c1] = self.data
        
        r0, r1, c0, c1 = out._shift_index(other.corners)
        
        other_data[r0:r1,c0:c1] = other.data
        
        return out, other_data
    
    def __getitem__(self, key):
        if type(key) == tuple:
            slc = []
            full = []
            for i, val in enumerate(key):
                if isinstance(val, slice):
                    f, s = self._shift_slice(val, i)
                    slc.append(s)
                    full.append(f)
                elif type(key) == int:
                    idx = self._get_idx(val, i)
                    if self.corners[i*2] <= idx < self.corners[i*2 + 1]:
                        slc.append(idx - self.corners[i*2])
                        full.append(False)
                    else:
                        slc.append(val)
                        full.append(True)
                else:
                    slc.append(val)
                    full.append(True)
            if any(full):
                return self.full()[key]
            else:
                return self.data[tuple(slc)]
        else:
            return self.full()[key]


from .poni import PONI, get_poni_dict, create_ai_from_dict
from .nzarrays import nzarray1d, nzarray2d
from .int_data import int_1d_data, int_2d_data
from .int_data_static import int_1d_data_static, int_2d_data_static

from ._containers import *



# -*- coding: utf-8 -*-
"""
Created on Mon Aug 26 14:21:58 2019

@author: walroth
"""
import copy
from threading import Condition

from pyFAI.azimuthalIntegrator import AzimuthalIntegrator
from pyFAI.containers import Integrate1dResult, Integrate2dResult
from pyFAI import units
import numpy as np

from xdart import utils
from xdart.utils.containers import PONI, int_1d_data, int_2d_data, create_ai_from_dict
from xdart.utils.containers import int_1d_data_static, int_2d_data_static

from icecream import ic; ic.configureOutput(prefix='', includeContext=True)


class EwaldArch():
    """Class for storing area detector data collected in
    X-ray diffraction experiments.

    Attributes:
        ai_args: dict, arguments passed to AzimuthalIntegrator
        arch_lock: Condition, threading lock used to ensure only one
            process can access data at a time
        file_lock: Condition, lock to ensure only one writer to
            data file
        idx: int, integer name of arch
        int_1d: int_1d_data/_static object from containers (for scanning/static detectors)
        int_2d: int_2d_data/_static object from containers (for scanning/static detectors)
        integrator: AzimuthalIntegrator object from pyFAI
        map_raw: numpy 2d array of the unprocessed image data
        bg_raw: numpy 2d array of the unprocessed image data for BG
        map_norm: float, normalization constant
        mask: numpy array of indeces to be masked in array.
        poni: poni data for integration
        poni_dict: poni_file information saved in dictionary
        scan_info: dict, information from any relevant motors and
            sensors
        static: bool, flag to specify if detector is static
        gi: bool, flag to specify if scattering geometry is grazing incidence
        th_mtr: str, the motor that controls sample rotation in gi mode
        tilt_angle: float, chi offset in gi geometry
        series_average: bool, flag to specify if series of images is averaged

    Methods:
        copy: create copy of arch
        get_mask: return mask array to feed into integrate1d
        integrate_1d: integrate the image data, results stored in
            int_1d_data
        integrate_2d: integrate the image data, results stored in
            int_2d_data
        load_from_h5: load data from hdf5 file
        save_to_h5: save data to hdf5 file
        set_integrator: set new integrator
        set_map_raw: replace raw data
        set_mask: replace mask data
        set_poni: replace poni object
        set_scan_info: replace scan_info
    """
    # pylint: disable=too-many-instance-attributes

    def __init__(self, idx=None, map_raw=None, poni=None, mask=None,
                 scan_info={}, ai_args={}, file_lock=Condition(),
                 static=False, poni_dict=None, bg_raw=0,
                 gi=False, th_mtr='th', tilt_angle=0,
                 series_average=False
                 ):
        # pylint: disable=too-many-arguments
        """idx: int, name of the arch.
        map_raw: numpy array, raw image data
        bg_raw: numpy array, raw image data for BG
        poni: PONI object, calibration data
        mask: None or numpy array, indices of pixels to mask
        scan_info: dict, metadata about scan
        ai_args: dict, args to be fed to azimuthalIntegrator constructor
        file_lock: Condition, lock for file access.
        """
        super(EwaldArch, self).__init__()
        self.idx = idx
        self.map_raw = map_raw
        self.bg_raw = bg_raw
        if poni is None:
            self.poni = PONI()
        else:
            self.poni = poni
        self.poni_dict = poni_dict
        if mask is None and map_raw is not None:
            self.mask = np.arange(map_raw.size)[map_raw.flatten() < 0]
        else:
            self.mask = mask
        self.scan_info = scan_info
        self.ai_args = ai_args
        self.file_lock = file_lock

        self.static = static
        self.gi = gi
        self.th_mtr = th_mtr
        self.tilt_angle = tilt_angle
        self.series_average = series_average

        self.integrator = self.setup_integrator()

        self.arch_lock = Condition()
        self.map_norm = 1

        if self.static:
            self.int_1d = int_1d_data_static()
            self.int_2d = int_2d_data_static()
        else:
            self.int_1d = int_1d_data()
            self.int_2d = int_2d_data()

    def setup_integrator(self):
        """Sets up integrator object"""
        # if self.poni_dict is not None:
        if self.static:
            if self.poni_dict is not None:
                integrator = create_ai_from_dict(self.poni_dict, self.gi)
            else:
                integrator = AzimuthalIntegrator()
        else:
            integrator = AzimuthalIntegrator(
                dist=self.poni.dist,
                poni1=self.poni.poni1,
                poni2=self.poni.poni2,
                rot1=self.poni.rot1,
                rot2=self.poni.rot2,
                rot3=self.poni.rot3,
                wavelength=self.poni.wavelength,
                detector=self.poni.detector,
                **self.ai_args
            )
        return integrator

    def reset(self):
        """Clears all data, resets to a default EwaldArch.
        """
        self.idx = None
        self.map_raw = None
        self.bg_raw = None
        self.poni = PONI()
        # self.poni_dict = None
        self.mask = None
        self.scan_info = {}
        self.integrator = self.setup_integrator()
        self.map_norm = 1
        if self.static:
            self.int_1d = int_1d_data_static()
            self.int_2d = int_2d_data_static()
        else:
            self.int_1d = int_1d_data()
            self.int_2d = int_2d_data()
            
    def get_mask(self, global_mask=None):
        if global_mask is not None:
            mask_idx = np.unique(np.append(self.mask, global_mask))
            # mask_idx.sort()
        else:
            mask_idx = self.mask
        mask = np.zeros(self.map_raw.size, dtype=int)
        try:
            mask[mask_idx] = 1
            return mask.reshape(self.map_raw.shape)
        except IndexError:
            print('Mask File Shape Mismatch')
            return mask

    def integrate_1d(self, numpoints=10000, radial_range=None,
                     monitor=None, unit=units.TTH_DEG, global_mask=None, **kwargs):
        """Wrapper for integrate1d method of AzimuthalIntegrator from pyFAI.
        Returns result and also stores the data in the int_1d object.

        args:
            numpoints: int, number of points in final array
            radial_range: tuple or list, lower and upper end of integration
            monitor: str, keyword for normalization counter in scan_info
            unit: pyFAI unit for integration, units.TTH_DEG, units.Q_A,
                '2th_deg', or 'q_A^-1'
            kwargs: other keywords to be passed to integrate1d, see pyFAI docs.

        returns:
            result: integrate1d result from pyFAI.
        """
        if (not self.static) and (radial_range is None):
            radial_range = [0, 180]

        with self.arch_lock:
            self.map_norm = 1
            if monitor is not None:
                if monitor.upper() in self.scan_info.keys():
                    self.map_norm = self.scan_info[monitor.upper()]
                elif monitor.lower() in self.scan_info.keys():
                    self.map_norm = self.scan_info[monitor.lower()]

            if self.mask is None:
                self.mask = np.arange(self.map_raw.size)[self.map_raw.flatten() < 0]

            if not self.gi:
                result = self.integrator.integrate1d(
                    # self.map_raw/self.map_norm, numpoints, unit=unit,
                    (self.map_raw-self.bg_raw)/self.map_norm, numpoints, unit=unit,
                    radial_range=radial_range, mask=self.get_mask(global_mask),
                    **kwargs
                )

                wavelength = self.poni.wavelength
                if self.static:
                    wavelength = self.integrator.wavelength
                self.int_1d.from_result(result, wavelength, unit=unit)
            else:
                pg_args = ['process', 'filename', 'correctSolidAngle', 'variance', 'error_model',
                           'dummy', 'delta_dummy', 'polarization_factor', 'dark', 'flat',
                           'method', 'safe', 'normalization_factor']
                pg_args = {k: v for (k, v) in kwargs.items() if k in pg_args}

                # incident angle in deg
                try:
                    self.integrator.incident_angle = float(self.th_mtr)
                except ValueError:
                    try:
                        self.integrator.incident_angle = self.scan_info[self.th_mtr]
                    except KeyError:
                        pass

                # ic(self.integrator.incident_angle)
                # tilt angle of sample in deg (misalignment in "chi")
                self.integrator.tilt_angle = self.tilt_angle

                Intensity, qAxis = self.integrator.integrate_1d(
                    # self.map_raw/self.map_norm, numpoints, unit='q_A^-1',
                    (self.map_raw-self.bg_raw)/self.map_norm, numpoints, unit='q_A^-1',
                    p0_range=radial_range, p1_range=kwargs['azimuth_range'],
                    mask=self.get_mask(global_mask), **pg_args
                )
                result = Integrate1dResult(qAxis, Intensity)
                self.int_1d.from_result(result, self.integrator.wavelength, unit='q_A^-1')

                # Get in-plane and out-of-plane contributions
                pg_args.pop('safe')
                i_qz, qz = self.integrator.profile_op_box(
                    (self.map_raw-self.bg_raw)/self.map_norm, numpoints, unit='q_A^-1',
                    ip_width=360.,
                    # p0_range=radial_range, p1_range=kwargs['azimuth_range'],
                    mask=self.get_mask(global_mask), **pg_args
                )
                self.int_1d.i_qz, self.int_1d.qz = i_qz, qz

                i_qxy, qxy = self.integrator.profile_ip_box(
                    (self.map_raw-self.bg_raw)/self.map_norm, numpoints, unit='q_A^-1',
                    op_width=360.,
                    # p0_range=radial_range, p1_range=kwargs['azimuth_range'],
                    mask=self.get_mask(global_mask), **pg_args
                )
                self.int_1d.i_qxy, self.int_1d.qxy = i_qxy, qxy

        # q = result.radial
        del result

    def integrate_2d(self, npt_rad=1000, npt_azim=1000, monitor=None,
                     radial_range=None, azimuth_range=None,
                     x_range=None, y_range=None,
                     unit=units.TTH_DEG, global_mask=None, **kwargs):
        """Wrapper for integrate2d method of AzimuthalIntegrator from pyFAI.
        Returns result and also stores the data in the int_2d object.

        args:
            npt_rad: int, number of points in radial dimension. If
                None, will take number from the shape of map_norm
            npt_azim: int, number of points in azimuthal dimension. If
                None, will take number from the shape of map_norm
            radial_range: tuple or list, lower and upper end of integration
            azimuth_range: tuple or list, lower and upper end of integration
                in azimuthal direction
            monitor: str, keyword for normalization counter in scan_info
            unit: pyFAI unit for integration, units.TTH_DEG, units.Q_A,
                '2th_deg', or 'q_A^-1'
            kwargs: other keywords to be passed to integrate2d, see pyFAI docs.

        returns:
            result: integrate2d result from pyFAI.
        """
        if (not self.static) and (radial_range is None):
            radial_range = [0, 180]
        if (not self.static) and (azimuth_range is None):
            azimuth_range = [-180, 180]

        with self.arch_lock:
            if monitor is not None:
                if self.static:
                    self.map_norm = 1
                else:
                    if monitor.upper() in self.scan_info.keys():
                        self.map_norm = self.scan_info[monitor.upper()]
                    elif monitor.lower() in self.scan_info.keys():
                        self.map_norm = self.scan_info[monitor.lower()]
                    else:
                        self.map_norm = 1

                    # self.map_norm = self.scan_info[monitor]

            if self.mask is None:
                self.mask = np.arange(self.map_raw.size)[self.map_raw.flatten() < 0]
            
            if npt_rad is None:
                npt_rad = self.map_raw.shape[0]
            
            if npt_azim is None:
                npt_azim = self.map_raw.shape[1]

            if not self.gi:
                result = self.integrator.integrate2d(
                    # self.map_raw/self.map_norm, npt_rad, npt_azim, unit=unit,
                    (self.map_raw-self.bg_raw)/self.map_norm, npt_rad, npt_azim, unit=unit,
                    mask=self.get_mask(global_mask), radial_range=radial_range,
                    azimuth_range=azimuth_range, **kwargs
                )
                wavelength = self.poni.wavelength
                if self.static:
                    wavelength = self.integrator.wavelength
                self.int_2d.from_result(result, wavelength, unit=unit)
            else:
                pg_args = ['filename', 'correctSolidAngle', 'variance', 'error_model',
                           'dummy', 'delta_dummy', 'polarization_factor', 'dark', 'flat',
                           'method', 'safe', 'normalization_factor']
                pg_args = {k: v for (k, v) in kwargs.items() if k in pg_args}

                # incident angle in deg
                try:
                    self.integrator.incident_angle = float(self.th_mtr)
                except ValueError:
                    try:
                        self.integrator.incident_angle = self.scan_info[self.th_mtr]
                    except KeyError:
                        pass

                # ic(self.integrator.incident_angle)

                # tilt angle of sample in deg (misalignment in "chi")
                self.integrator.tilt_angle = self.tilt_angle

                if unit == '2th_deg':
                    radial_range = self.convert_radial_range(radial_range, self.integrator.wavelength)

                # Transform to polar (Q-Chi) coordinates
                i_qchi, Q, Chi = self.integrator.transform_image(
                    self.map_raw-self.bg_raw, process='polar', npt=(npt_rad, npt_azim),
                    x_range=radial_range, y_range=azimuth_range, unit='q_A^-1',
                    mask=self.get_mask(global_mask), all=False, **pg_args)
                result = Integrate2dResult(i_qchi, Q, Chi)

                # Transform to reciprocal (Qz-Qxy) coordinates
                i_QxyQz, qxy, qz = self.integrator.transform_image(
                    self.map_raw-self.bg_raw, process='reciprocal', npt=(npt_rad, npt_azim),
                    x_range=x_range, y_range=y_range, unit='q_A^-1',
                    mask=self.get_mask(), all=False, **pg_args)

                self.int_2d.from_result(result, self.integrator.wavelength, unit=unit,
                                        i_QxyQz=np.flipud(i_QxyQz), qz=qz, qxy=qxy)

    def set_integrator(self, **args):
        """Sets AzimuthalIntegrator with new arguments.

        args:
            args: see pyFAI for acceptable arguments for the integrator
                constructor.

        returns:
            None
        """

        with self.arch_lock:
            self.ai_args = args
            self.integrator = self.setup_integrator()

    def set_map_raw(self, new_data):
        with self.arch_lock:
            self.map_raw = new_data
            if self.mask is None:
                self.mask = np.arange(new_data.size)[new_data.flatten() < 0]

    def set_poni(self, new_data):
        with self.arch_lock:
            self.poni = new_data

    def set_mask(self, new_data):
        with self.arch_lock:
            self.mask = new_data

    def set_scan_info(self, new_data):
        with self.arch_lock:
            self.scan_info = new_data

    @staticmethod
    def convert_radial_range(radial_range, wavelength):
        """Convert radial range from Q (AA-1) to 2Th (deg)"""
        if radial_range is None:
            return None

        radial_range = np.asarray(radial_range)
        radial_range = (4 * np.pi / (wavelength * 1e10)) *\
            np.sin(np.radians(radial_range / 2))

        return radial_range

    def save_to_h5(self, file, compression='lzf'):
        """Saves data to hdf5 file using h5py as backend.

        args:
            file: h5py group or file object.

        returns:
            None
        """
        if self.static:
            compression = None
        with self.file_lock:
            if str(self.idx) in file:
                grp = file[str(self.idx)]
            else:
                grp = file.create_group(str(self.idx))
            grp.attrs['type'] = 'EwaldArch'
            lst_attr = [
                "map_raw", "mask", "map_norm", "scan_info", "ai_args",
                "gi", "static", "poni_dict", "bg_raw"
            ]
            utils.attributes_to_h5(self, grp, lst_attr,
                                   compression=compression)
            if 'int_1d' not in grp:
                grp.create_group('int_1d')
            self.int_1d.to_hdf5(grp['int_1d'], compression)
            if 'int_2d' not in grp:
                grp.create_group('int_2d')
            self.int_2d.to_hdf5(grp['int_2d'], compression)
            if 'poni' not in grp:
                grp.create_group('poni')
            utils.dict_to_h5(self.poni.to_dict(), grp, 'poni')

    def load_from_h5(self, file, load_2d=True):
        """Loads data from hdf5 file and sets attributes.

        args:
            file: h5py file or group object
        """
        with self.file_lock:
            with self.arch_lock:
                if str(self.idx) not in file:
                    print("No data can be found")
                else:
                    grp = file[str(self.idx)]
                    if 'type' in grp.attrs:
                        if grp.attrs['type'] == 'EwaldArch':
                            lst_attr = [
                                "map_raw", "mask", "map_norm", "scan_info", "ai_args",
                                "gi", "static", "poni_dict", "bg_raw"
                            ]
                            utils.h5_to_attributes(self, grp, lst_attr)
                            self.int_1d.from_hdf5(grp['int_1d'])
                            if load_2d:
                                self.int_2d.from_hdf5(grp['int_2d'])
                            self.poni = PONI.from_yamdict(
                                utils.h5_to_dict(grp['poni'])
                            )

                            if self.poni_dict is not None:
                                self.integrator = create_ai_from_dict(self.poni_dict)
                            else:
                                self.integrator = AzimuthalIntegrator(
                                    dist=self.poni.dist,
                                    poni1=self.poni.poni1,
                                    poni2=self.poni.poni2,
                                    rot1=self.poni.rot1,
                                    rot2=self.poni.rot2,
                                    rot3=self.poni.rot3,
                                    wavelength=self.poni.wavelength,
                                    detector=self.poni.detector,
                                    **self.ai_args
                                )

    def copy(self, include_2d=True):
        """Returns a copy of self.
        """
        arch_copy = EwaldArch(
            copy.deepcopy(self.idx), None,
            copy.deepcopy(self.poni), None,
            copy.deepcopy(self.scan_info), copy.deepcopy(self.ai_args),
            self.file_lock, poni_dict=copy.deepcopy(self.poni_dict),
            static=copy.deepcopy(self.static), gi=copy.deepcopy(self.gi),
            th_mtr=copy.deepcopy(self.th_mtr),
            series_average=copy.deepcopy(self.series_average)
        )
        arch_copy.integrator = copy.deepcopy(self.integrator)
        arch_copy.arch_lock = Condition()
        arch_copy.int_1d = copy.deepcopy(self.int_1d)
        if include_2d:
            arch_copy.map_raw = copy.deepcopy(self.map_raw)
            arch_copy.bg_raw = copy.deepcopy(self.bg_raw)
            arch_copy.mask = copy.deepcopy(self.mask),
            arch_copy.map_norm = copy.deepcopy(self.map_norm)
            arch_copy.int_2d = copy.deepcopy(self.int_2d)

        return arch_copy



# -*- coding: utf-8 -*-
"""
@author: thampy
"""

# Standard library imports
import os
import sys
import subprocess

import fabio
from xdart.utils.pyFAI_binaries import pyFAI_drawmask_main
from xdart.utils.pyFAI_binaries import MaskImageWidgetXdart
from pyFAI.app.drawmask import postProcessId21

# Qt imports
import pyqtgraph as pg
from pyqtgraph import Qt
from pyqtgraph.parametertree import Parameter

# This module imports
if sys.platform == 'win32':
    from .ui.integratorUI_windows import Ui_Form
else:
    from .ui.integratorUI_mac import Ui_Form

from .sphere_threads import integratorThread

_translate = Qt.QtCore.QCoreApplication.translate
QFileDialog = Qt.QtWidgets.QFileDialog

AA_inv = u'\u212B\u207B\u00B9'
Th = u'\u03B8'
Deg = u'\u00B0'
Units = [f"Q ({AA_inv})", f"2{Th} ({Deg})"]
Units_dict = {Units[0]: 'q_A^-1', Units[1]: '2th_deg'}
# Units_dict_inv = {'q_A^-1': Units[0], '2th_deg': Units[1]}
Units_dict_inv = {'q_A^-1': 0, '2th_deg': 1}

params = [
    {'name': 'Default', 'type': 'group', 'children': [
            {'name': 'Integrate 1D', 'type': 'group', 'children': [
                # {'name': 'numpoints', 'type': 'int', 'value': 1000, 'visible': False},
                # {'name': 'unit', 'type': 'list', 'values': {
                #     # "2" + u"\u03B8": '2th_deg', "q (A-1)": 'q_A^-1'
                #     Units[0]: 'q_A^-1', Units[1]: '2th_deg',
                #     }, 'value': 'q_A^-1', 'visible':False
                #  },  # 2th_deg
                # {'name': 'radial_range', 'type': 'group', 'children': [
                #     {'name': 'Low', 'type': 'float', 'value': 0.0, 'visible': False},
                #     {'name': 'High', 'type': 'float', 'value': 60.0, 'visible': False},
                #     {'name': 'Auto', 'type': 'bool', 'value': True, 'visible': False},
                #     ], 'visible': False,
                #  },
                # {'name': 'azimuth_range', 'type': 'group', 'children': [
                #         {'name': 'Low', 'type': 'float', 'value': -180.0, 'visible': False},
                #         {'name': 'High', 'type': 'float', 'value': 180.0, 'visible': False},
                #         {'name': 'Auto', 'type': 'bool', 'value': True, 'visible': False},
                #     ], 'visible': False,
                #  },
                # {'name': 'monitor', 'type': 'str', 'value': 'I0'},
                {'name': 'correctSolidAngle', 'type': 'bool', 'value': True},
                {'name': 'dummy', 'type': 'float', 'value': -1.0},
                {'name': 'delta_dummy', 'type': 'float', 'value': 0.0},
                {'name': 'Apply polarization factor', 'type': 'bool', 'value': False},
                {'name': 'polarization_factor', 'type': 'float', 'value': 0,
                    'limits': (-1, 1)},
                {'name': 'method', 'type': 'list', 'values': [
                        "numpy", "cython", "BBox", "splitpixel", "lut", "csr",
                        "nosplit_csr", "full_csr", "lut_ocl", "csr_ocl"
                    ], 'value':'csr'},
                {'name': 'safe', 'type': 'bool', 'value': True},
                # {'name': 'block_size', 'type': 'int', 'value': 32},
                # {'name': 'profile', 'type': 'bool', 'value': False},
                ]
             },
            {'name': 'Integrate 2D', 'type': 'group', 'children': [
                # {'name': 'npt_rad', 'type': 'int', 'value': 1000, 'visible': False},
                # {'name': 'npt_azim', 'type': 'int', 'value': 1000, 'visible': False},
                # {'name': 'unit', 'type': 'list', 'values': {
                #     Units[0]: 'q_A^-1', Units[1]: '2th_deg',
                #     # "2" + u"\u03B8": '2th_deg', "q (A-1)": 'q_A^-1'
                #     }, 'value': 'q_A^-1', 'visible': False},
                # {'name': 'radial_range', 'type': 'group', 'children': [
                #         {'name': 'Low', 'type': 'float', 'value': 0.0},
                #         {'name': 'High', 'type': 'float', 'value': 180.0},
                #         {'name': 'Auto', 'type': 'bool', 'value': True},
                #     ], 'visible': False,
                #  },
                # {'name': 'azimuth_range', 'type': 'group', 'children': [
                #         {'name': 'Low', 'type': 'float', 'value': -180.0},
                #         {'name': 'High', 'type': 'float', 'value': 180.0},
                #         {'name': 'Auto', 'type': 'bool', 'value': True,
                #          'visible': True},
                #     ], 'visible': False,
                #  },
                # {'name': 'monitor', 'type': 'str', 'value': 'None'},
                {'name': 'correctSolidAngle', 'type': 'bool', 'value': True},
                {'name': 'dummy', 'type': 'float', 'value': -1.0},
                {'name': 'delta_dummy', 'type': 'float', 'value': 0.0},
                {'name': 'Apply polarization factor', 'type': 'bool', 'value': False},
                {'name': 'polarization_factor', 'type': 'float', 'value': 0,
                    'limits': (-1, 1)},
                {'name': 'method', 'type': 'list', 'values': [
                        "numpy", "cython", "BBox", "splitpixel", "lut",
                        "csr", "lut_ocl", "csr_ocl"
                    ], 'value':'csr'},
                {'name': 'safe', 'type': 'bool', 'value': True}
                ]
             }
        ]

     },
]


class integratorTree(Qt.QtWidgets.QWidget):
    """Widget for controlling integration of loaded data. Presents basic
    parameters to the user in easy to control widgets, and also
    launches menus for more advanced options.

    attributes:
        advancedWidget1D, advancedWidget2D: advancedParameters, pop up
            windows with advanced parameters
        azimuthalRange2D, radialRange1D, radialRange2D: rangeWidget,
            widgets which control the integration ranges for 1D and 2D
            integration
        bai_1d_pars, bai_2d_pars: pyqtgraph parameters, children of
            main parameters attribute that hold parameters related to
            1D and 2D integration
        mg_1d_pars, mg_2d_pars: unused, hold paramters for multigeometry
            integration
        mg_pars: unused, holds parameters for setting up multigeometry
        ui: Ui_Form from qtdesigner

    methods:
        get_args: Gets requested parameters and converts them to args
            in EwaldSphere object.
        setEnabled: Enables integration and parameter modification.
        update: Grabs args from EwaldSphere object and sets all params
            to match.
    """
    def __init__(self, sphere, arch, file_lock,
                 arches, arch_ids, data_1d, data_2d, parent=None):
        super().__init__(parent)
        self.ui = Ui_Form()
        self.ui.setupUi(self)
        self.sphere = sphere
        self.arch = arch
        self.file_lock = file_lock
        self.arches = arches
        self.arch_ids = arch_ids
        self.data_1d = data_1d
        self.data_2d = data_2d
        self.parameters = Parameter.create(
            name='integrator', type='group', children=params
        )
        self.bai_1d_pars = self.parameters.child('Default', 'Integrate 1D')
        self.bai_2d_pars = self.parameters.child('Default', 'Integrate 2D')
        self.mask_window = None

        # UI adjustments
        _translate = Qt.QtCore.QCoreApplication.translate
        self.ui.unit_1D.setItemText(0, _translate("Form", Units[0]))
        self.ui.unit_1D.setItemText(1, _translate("Form", Units[1]))
        self.ui.label_azim_1D.setText(f"Chi ({Deg})")

        self.ui.unit_2D.setItemText(0, _translate("Form", Units[0]))
        self.ui.unit_2D.setItemText(1, _translate("Form", Units[1]))
        self.ui.label_azim_2D.setText(f"Chi ({Deg})")

        # Set constraints on range inputs
        self._validate_ranges()

        # Set AutoRange Flags
        self.radial_autoRange_1D = self.ui.radial_autoRange_1D.isChecked()
        self.azim_autoRange_1D = self.ui.azim_autoRange_1D.isChecked()
        self.radial_autoRange_2D = self.ui.radial_autoRange_2D.isChecked()
        self.azim_autoRange_2D = self.ui.azim_autoRange_2D.isChecked()

        # Setup advanced parameters tree widget
        self.advancedWidget1D = advancedParameters(self.bai_1d_pars, 'bai_1d')
        self.advancedWidget2D = advancedParameters(self.bai_2d_pars, 'bai_2d')

        # Connect input parameter value signals
        self._connect_inp_signals()

        # Connect integrate and advanced button signals
        self.ui.advanced1D.clicked.connect(self.advancedWidget1D.show)
        self.ui.advanced2D.clicked.connect(self.advancedWidget2D.show)

        self.ui.integrate1D.clicked.connect(self.bai_1d)
        self.ui.integrate2D.clicked.connect(self.bai_2d)

        self.integrator_thread = integratorThread(
            self.sphere, self.arch, self.file_lock,
            self.arches, self.arch_ids, self.data_1d, self.data_2d
        )

        # Connect Calibrate and Mask Buttons
        self.ui.pyfai_calib.clicked.connect(self.run_pyfai_calib)
        self.ui.get_mask.clicked.connect(self.run_pyfai_drawmask)

        self.setEnabled()
        # self.set_image_units()

        # Connect Axis 2D signal
        self.ui.axis2D.currentIndexChanged.connect(self._update_axes)

    def update(self):
        """Grabs args from sphere and uses _sync_ranges and
        _update_params private methods to update.

        args:
            sphere: EwaldSphere, object to get args from.
        """
        self._update_params()

    def setEnabled(self, enable=True):
        """Overrides parent class method. Ensures appropriate child
        widgets are enabled.

        args:
            enable: bool, If True widgets are enabled. If False
                they are disabled.
        """
        self.ui.frame1D.setEnabled(enable)
        self.ui.frame2D.setEnabled(enable)
        self.advancedWidget1D.setEnabled(enable)
        self.advancedWidget2D.setEnabled(enable)

        if self.radial_autoRange_1D:
            self.ui.radial_low_1D.setEnabled(False)
            self.ui.radial_high_1D.setEnabled(False)

        if self.azim_autoRange_1D:
            self.ui.azim_low_1D.setEnabled(False)
            self.ui.azim_high_1D.setEnabled(False)

        if self.radial_autoRange_2D:
            self.ui.radial_low_2D.setEnabled(False)
            self.ui.radial_high_2D.setEnabled(False)

        if self.azim_autoRange_2D:
            self.ui.azim_low_2D.setEnabled(False)
            self.ui.azim_high_2D.setEnabled(False)

        # self.ui.integrate1D.setEnabled(False)
        # self.ui.integrate2D.setEnabled(False)

    def update_radial_autoRange_1D(self):
        """Disable/Enable radial 1D widget if auto range is un/selected
        """
        self.radial_autoRange_1D = self.ui.radial_autoRange_1D.isChecked()
        # self.setEnabled()

        self.bai_1d_pars.child('radial_range', 'Auto').setValue(
            self.radial_autoRange_1D)
        if self.radial_autoRange_1D:
            self.sphere.bai_1d_args['radial_range'] = None
            self.ui.radial_low_1D.setEnabled(False)
            self.ui.radial_high_1D.setEnabled(False)
        else:
            self.ui.radial_low_1D.setEnabled(True)
            self.ui.radial_high_1D.setEnabled(True)

        # print(f'integrator > update_radial_autoRange_1D: sphere_bai_1d = {self.sphere.bai_1d_args}')

    def update_azim_autoRange_1D(self):
        """Disable/Enable azim 1D widget if auto range is un/selected
        """
        self.azim_autoRange_1D = self.ui.azim_autoRange_1D.isChecked()

        self.bai_1d_pars.child('azimuth_range', 'Auto').setValue(
            self.azim_autoRange_1D)

        if self.azim_autoRange_1D:
            self.sphere.bai_1d_args['azimuth_range'] = None
            self.ui.azim_low_1D.setEnabled(False)
            self.ui.azim_high_1D.setEnabled(False)
        else:
            self.ui.azim_low_1D.setEnabled(True)
            self.ui.azim_high_1D.setEnabled(True)

        # print(f"integrator > update_azim_autoRange_1D: radial_range {self.sphere.bai_1d_args['radial_range']}")

    def update_radial_autoRange_2D(self):
        """Disable/Enable radial 2D widget if auto range is un/selected
        """
        #ic()
        self.radial_autoRange_2D = self.ui.radial_autoRange_2D.isChecked()

        self.bai_2d_pars.child('radial_range', 'Auto').setValue(
            self.radial_autoRange_2D)

        if self.radial_autoRange_2D:
            self.sphere.bai_2d_args['radial_range'] = None
            self.ui.radial_low_2D.setEnabled(False)
            self.ui.radial_high_2D.setEnabled(False)
        else:
            self.ui.radial_low_2D.setEnabled(True)
            self.ui.radial_high_2D.setEnabled(True)

        # print(f"integrator > update_radial_autoRange_2D: radial_range {self.sphere.bai_2d_args['radial_range']}")

    def update_azim_autoRange_2D(self):
        """Disable/Enable radial 2D widget if auto range is un/selected
        """
        #ic()
        self.azim_autoRange_2D = self.ui.azim_autoRange_2D.isChecked()

        self.bai_2d_pars.child('azimuth_range', 'Auto').setValue(
            self.azim_autoRange_2D)

        if self.azim_autoRange_2D:
            self.sphere.bai_2d_args['azimuth_range'] = None
            self.ui.azim_low_2D.setEnabled(False)
            self.ui.azim_high_2D.setEnabled(False)
        else:
            self.ui.azim_low_2D.setEnabled(True)
            self.ui.azim_high_2D.setEnabled(True)

        # print(f"integrator > update_azim_autoRange_2D: azimuth_range = {self.sphere.bai_2d_args['azimuth_range']}")

    def _validate_ranges(self):
        #ic()
        self.ui.npts_1D.setValidator(Qt.QtGui.QIntValidator(0, 50000))
        self.ui.npts_radial_2D.setValidator(Qt.QtGui.QIntValidator(0, 50000))
        self.ui.npts_azim_2D.setValidator(Qt.QtGui.QIntValidator(0, 50000))

        minmax = (0, 50)
        if self.ui.unit_1D.currentIndex() == 1:
            minmax = (-180, 180)
        self.ui.radial_low_1D.setValidator(Qt.QtGui.QDoubleValidator(minmax[0], minmax[1], 2))
        self.ui.radial_high_1D.setValidator(Qt.QtGui.QDoubleValidator(minmax[0], minmax[1], 2))

        minmax = (0, 50)
        if self.ui.unit_2D.currentIndex() == 1:
            minmax = (-180, 180)
        self.ui.radial_low_2D.setValidator(Qt.QtGui.QDoubleValidator(minmax[0], minmax[1], 2))
        self.ui.radial_high_2D.setValidator(Qt.QtGui.QDoubleValidator(minmax[0], minmax[1], 2))

        self.ui.azim_low_1D.setValidator(Qt.QtGui.QDoubleValidator(-180, 180, 2))
        self.ui.azim_high_1D.setValidator(Qt.QtGui.QDoubleValidator(-180, 180, 2))
        self.ui.azim_low_2D.setValidator(Qt.QtGui.QDoubleValidator(-180, 180, 2))
        self.ui.azim_high_2D.setValidator(Qt.QtGui.QDoubleValidator(-180, 180, 2))

    def _update_params(self):
        """Grabs args from sphere and syncs parameters with them.

        args:
            sphere: EwaldSphere, object to get args from.
        """
        #ic()

        self._disconnect_inp_signals()
        with self.sphere.sphere_lock:
            self._args_to_params(self.sphere.bai_1d_args, self.bai_1d_pars, dim='1D')
            self._args_to_params(self.sphere.bai_2d_args, self.bai_2d_pars, dim='2D')
        self._connect_inp_signals()

    def get_args(self, key):
        """Updates sphere with all parameters held in integrator.

        args:
            sphere: EwaldSphere, object to update
            key: str, which args to update.
        """
        #ic()
        with self.sphere.sphere_lock:
            if key == 'bai_1d':
                #ic('bai_1D_args before', self.sphere.bai_1d_args)
                self._get_npts_1D()
                self._get_unit_1D()
                self._get_radial_range_1D()
                self._get_azim_range_1D()
                self._params_to_args(self.sphere.bai_1d_args, self.bai_1d_pars)
                #ic('bai_1D_args after', self.sphere.bai_1d_args)

            elif key == 'bai_2d':
                #ic('bai_2D_args before', self.sphere.bai_2d_args)
                self._get_npts_radial_2D()
                self._get_npts_azim_2D()
                self._get_unit_2D()
                self._get_radial_range_2D()
                self._get_azim_range_2D()
                self._params_to_args(self.sphere.bai_2d_args, self.bai_2d_pars)
                #ic('bai_2D_args after', self.sphere.bai_2d_args)

    def _args_to_params(self, args, tree, dim='1D'):
        """Takes in args dictionary and sets all parameters in tree
        to match the args.

        args:
            args: dict, values to use for updating tree
            tree: pyqtgraph Parameter, parameters to update
        """
        #ic()

        #ic(args)
        if len(args) == 0:
            return

        with tree.treeChangeBlocker():
            for key, val in args.items():
                if key == 'radial_range':
                    if dim == '1D':
                        self._set_radial_range_1D()
                    else:
                        self._set_radial_range_2D()
                elif key == 'azimuth_range':
                    if dim == '1D':
                        self._set_azim_range_1D()
                    else:
                        self._set_azim_range_2D()
                elif key == 'unit':
                    if dim == '1D':
                        self._set_unit_1D()
                    else:
                        self._set_unit_2D()
                elif key == 'numpoints':
                    self._set_npts_1D()
                elif key == 'npt_rad':
                    self._set_npts_radial_2D()
                elif key == 'npt_azim':
                    self._set_npts_azim_2D()
                elif key == 'polarization_factor':
                    if val is None:
                        tree.child('Apply polarization factor').setValue(True)
                    else:
                        tree.child('Apply polarization factor').setValue(True)
                        tree.child(key).setValue(val)
                else:
                    try:
                        child = tree.child(key)
                    except:
                        # No specific error thrown for missing child
                        child = None
                    if child is not None:
                        if val is None:
                            child.setValue('None')
                        else:
                            child.setValue(val)

    def _params_to_args(self, args, tree):
        """Sets all values in args to match tree.

        args:
            args: dict, values to be updates
            tree: pyqtgraph Parameter, parameters used to update args
        """
        #ic()

        #ic(tree, args)
        for child in tree.children():
            # print(f'integrator > _params_to_args: child.name, value = {child.name()}, {child.value()}')
            if 'range' in child.name():
                if child.child("Auto").value():
                    args[child.name()] = None
                else:
                    args[child.name()] = [child.child("Low").value(),
                                          child.child("High").value()]
            elif child.name() == 'polarization_factor':
                pass
            elif child.name() == 'Apply polarization factor':
                if child.value():
                    args['polarization_factor'] = \
                        tree.child('polarization_factor').value()
                else:
                    args['polarization_factor'] = None
            else:
                val = child.value()
                if val == 'None':
                    args[child.name()] = None
                else:
                    args[child.name()] = val

    def _get_radial_range_1D(self):
        """Sets Sphere 1D radial range in bai_1d_args from UI values"""
        #ic()

        auto = self.ui.radial_autoRange_1D.isChecked()
        self.radial_autoRange_1D = auto

        _range = None
        if not auto:
            _range = self._get_valid_range(self.ui.radial_low_1D,
                                           self.ui.radial_high_1D)
        self.sphere.bai_1d_args['radial_range'] = _range

        self.ui.radial_low_1D.setEnabled(not auto)
        self.ui.radial_high_1D.setEnabled(not auto)
        #ic(_range)

    def _set_radial_range_1D(self):
        """Sets UI values from Sphere 1D radial range in bai_1d_args"""
        #ic()

        self._disconnect_radial_range_1D_signals()

        _range = self.sphere.bai_1d_args['radial_range']
        #ic(self.sphere.bai_1d_args, _range)
        if _range is None:
            self.ui.radial_autoRange_1D.setChecked(True)
            auto = True
        else:
            self.ui.radial_low_1D.setText(str(_range[0]))
            self.ui.radial_high_1D.setText(str(_range[0]))
            auto = False

        self.radial_autoRange_1D = auto
        self.ui.radial_low_1D.setEnabled(not auto)
        self.ui.radial_high_1D.setEnabled(not auto)

        self._connect_radial_range_1D_signals()

    def _get_azim_range_1D(self):
        """Sets Sphere 1D azimuth range in bai_1d_args from UI values"""
        #ic()

        auto = self.ui.azim_autoRange_1D.isChecked()
        self.azim_autoRange_1D = auto

        _range = None
        if not auto:
            _range = self._get_valid_range(self.ui.azim_low_1D,
                                           self.ui.azim_high_1D)
        self.sphere.bai_1d_args['azimuth_range'] = _range

        self.ui.azim_low_1D.setEnabled(not auto)
        self.ui.azim_high_1D.setEnabled(not auto)
        #ic(_range)

    def _set_azim_range_1D(self):
        """Sets UI values from Sphere 1D aazimuth range in bai_1d_args"""
        #ic()

        self._disconnect_azim_range_1D_signals()

        _range = self.sphere.bai_1d_args['azimuth_range']
        #ic(self.sphere.bai_1d_args, _range)
        if _range is None:
            self.ui.azim_autoRange_1D.setChecked(True)
            auto = True
        else:
            self.ui.azim_low_1D.setText(str(_range[0]))
            self.ui.azim_high_1D.setText(str(_range[0]))
            auto = False

        self.azim_autoRange_1D = auto
        self.ui.azim_low_1D.setEnabled(not auto)
        self.ui.azim_high_1D.setEnabled(not auto)

        self._connect_azim_range_1D_signals()

    def _get_radial_range_2D(self):
        """Sets Sphere 2D radial range in bai_1d_args from UI values"""
        #ic()

        auto = self.ui.radial_autoRange_2D.isChecked()
        self.radial_autoRange_2D = auto

        _range = None
        if not auto:
            _range = self._get_valid_range(self.ui.radial_low_2D,
                                           self.ui.radial_high_2D)

        if self.ui.axis2D.currentIndex() == 1:
            self.sphere.bai_2d_args['x_range'] = _range
        else:
            self.sphere.bai_2d_args['radial_range'] = _range

        self.ui.radial_low_2D.setEnabled(not auto)
        self.ui.radial_high_2D.setEnabled(not auto)
        #ic(_range)

    def _set_radial_range_2D(self):
        """Sets UI values from Sphere 2D radial range in bai_2d_args"""
        #ic()

        self._disconnect_radial_range_2D_signals()

        _range = self.sphere.bai_2d_args['radial_range']
        #ic(self.sphere.bai_2d_args, _range)
        if _range is None:
            self.ui.radial_autoRange_2D.setChecked(True)
            auto = True
        else:
            self.ui.radial_low_2D.setText(str(_range[0]))
            self.ui.radial_high_2D.setText(str(_range[0]))
            auto = False

        self.radial_autoRange_2D = auto
        self.ui.radial_low_2D.setEnabled(not auto)
        self.ui.radial_high_2D.setEnabled(not auto)

        self._connect_radial_range_2D_signals()

    def _get_azim_range_2D(self):
        """Sets Sphere 2D azimuth range in bai_1d_args from UI values"""
        #ic()

        auto = self.ui.azim_autoRange_2D.isChecked()
        self.azim_autoRange_2D = auto

        _range = None
        if not auto:
            _range = self._get_valid_range(self.ui.azim_low_2D,
                                           self.ui.azim_high_2D)

        if self.ui.axis2D.currentIndex() == 1:
            self.sphere.bai_2d_args['y_range'] = _range
        else:
            self.sphere.bai_2d_args['azimuth_range'] = _range

        self.ui.azim_low_2D.setEnabled(not auto)
        self.ui.azim_high_2D.setEnabled(not auto)
        #ic(_range)

    def _set_azim_range_2D(self):
        """Sets UI values from Sphere 1D aazimuth range in bai_1d_args"""
        #ic()

        self._disconnect_azim_range_2D_signals()

        _range = self.sphere.bai_2d_args['azimuth_range']
        #ic(self.sphere.bai_2d_args, _range)
        if _range is None:
            self.ui.azim_autoRange_2D.setChecked(True)
            auto = True
        else:
            self.ui.azim_low_2D.setText(str(_range[0]))
            self.ui.azim_high_2D.setText(str(_range[0]))
            auto = False

        self.azim_autoRange_1D = auto
        self.ui.azim_low_2D.setEnabled(not auto)
        self.ui.azim_high_2D.setEnabled(not auto)

        self._connect_azim_range_2D_signals()

    @staticmethod
    def _get_valid_range(low, high):
        """Validate range to return float values"""

        low, high = low.text(), high.text()
        try:
            low = float(low)
        except ValueError:
            low = 0.
        try:
            high = float(high)
        except ValueError:
            high = 0.
        return [low, high]

    def _get_unit_1D(self):
        #ic()

        val = self.ui.unit_1D.currentText()
        #ic(val, Units_dict[val])
        self.sphere.bai_1d_args['unit'] = Units_dict[val]
        self._validate_ranges()

    def _set_unit_1D(self):
        #ic()

        self.ui.unit_1D.currentTextChanged.disconnect(self._get_unit_1D)
        val = self.sphere.bai_1d_args['unit']
        #ic(val, type(val), Units_dict_inv[val])
        self.ui.unit_1D.setCurrentIndex(Units_dict_inv[val])
        self.ui.unit_1D.currentTextChanged.connect(self._get_unit_1D)

    def _get_unit_2D(self):
        #ic()

        val = self.ui.unit_2D.currentText()
        #ic(val, Units_dict[val])
        self.sphere.bai_2d_args['unit'] = Units_dict[val]
        self._validate_ranges()

    def _set_unit_2D(self):
        #ic()

        self.ui.unit_2D.currentTextChanged.disconnect(self._get_unit_2D)
        val = self.sphere.bai_2d_args['unit']
        #ic(val, type(val), Units_dict_inv[val])
        self.ui.unit_2D.setCurrentIndex(Units_dict_inv[val])
        # self.ui.unit_2D.setCurrentText(Units_dict_inv[val])
        self.ui.unit_2D.currentTextChanged.connect(self._get_unit_2D)

    def _get_npts_1D(self):
        #ic()

        val = self.ui.npts_1D.text()
        val = 500 if (not val) else int(val)
        #ic(val)
        self.sphere.bai_1d_args['numpoints'] = val

    def _set_npts_1D(self):
        #ic()

        self.ui.npts_1D.textChanged.disconnect(self._get_npts_1D)
        val = str(self.sphere.bai_1d_args['numpoints'])
        #ic(val, type(val))
        self.ui.npts_1D.setText(val)
        self.ui.npts_1D.textChanged.connect(self._get_npts_1D)

    def _get_npts_radial_2D(self):
        #ic()

        val = self.ui.npts_radial_2D.text()
        val = 500 if (not val) else int(val)
        #ic(val)
        self.sphere.bai_2d_args['npt_rad'] = val

    def _set_npts_radial_2D(self):
        #ic()

        self.ui.npts_radial_2D.textChanged.disconnect(self._get_npts_radial_2D)
        val = str(self.sphere.bai_2d_args['npt_rad'])
        #ic(val, type(val))
        self.ui.npts_radial_2D.setText(val)
        self.ui.npts_radial_2D.textChanged.connect(self._get_npts_radial_2D)

    def _get_npts_azim_2D(self):
        #ic()

        val = self.ui.npts_azim_2D.text()
        val = 500 if (not val) else int(val)
        #ic(val)
        self.sphere.bai_2d_args['npt_azim'] = val

    def _set_npts_azim_2D(self):
        #ic()

        self.ui.npts_azim_2D.textChanged.disconnect(self._get_npts_azim_2D)
        val = str(self.sphere.bai_2d_args['npt_azim'])
        #ic(val, type(val))
        self.ui.npts_azim_2D.setText(val)
        self.ui.npts_azim_2D.textChanged.connect(self._get_npts_azim_2D)

    def _connect_inp_signals(self):
        """Connect signals for all input sphere bai parameters"""
        #ic()

        # Connect points and units signals
        self.ui.npts_1D.textChanged.connect(self._get_npts_1D)
        self.ui.unit_1D.currentTextChanged.connect(self._get_unit_1D)

        self.ui.npts_radial_2D.textChanged.connect(self._get_npts_radial_2D)
        self.ui.npts_azim_2D.textChanged.connect(self._get_npts_azim_2D)
        self.ui.unit_2D.currentTextChanged.connect(self._get_unit_2D)

        # Connect range signals
        self._connect_radial_range_1D_signals()
        self._connect_azim_range_1D_signals()
        self._connect_radial_range_2D_signals()
        self._connect_azim_range_2D_signals()

        # Connect advanced parameters signals
        self.advancedWidget1D.sigUpdateArgs.connect(self.get_args)
        self.advancedWidget2D.sigUpdateArgs.connect(self.get_args)

    def _disconnect_inp_signals(self):
        """Disconnect signals for all input sphere bai parameters"""
        #ic()

        # Disconnect points and units signals
        self.ui.npts_1D.textChanged.disconnect(self._get_npts_1D)
        self.ui.unit_1D.currentTextChanged.disconnect(self._get_unit_1D)
        self.ui.npts_radial_2D.textChanged.disconnect(self._get_npts_radial_2D)
        self.ui.npts_azim_2D.textChanged.disconnect(self._get_npts_azim_2D)
        self.ui.unit_2D.currentTextChanged.disconnect(self._get_unit_2D)

        # Disconnect range signals
        self._disconnect_radial_range_1D_signals()
        self._disconnect_azim_range_1D_signals()
        self._disconnect_radial_range_2D_signals()
        self._disconnect_azim_range_2D_signals()

        # Disconnect advanced parameters signals
        self.advancedWidget1D.sigUpdateArgs.disconnect(self.get_args)
        self.advancedWidget2D.sigUpdateArgs.disconnect(self.get_args)

    def _connect_radial_range_1D_signals(self):
        """Connect signals for radial 1D range"""
        self.ui.radial_low_1D.textChanged.connect(self._get_radial_range_1D)
        self.ui.radial_high_1D.textChanged.connect(self._get_radial_range_1D)
        self.ui.radial_autoRange_1D.stateChanged.connect(self._get_radial_range_1D)

    def _disconnect_radial_range_1D_signals(self):
        """Disconnect signals for radial 1D range"""
        self.ui.radial_low_1D.textChanged.disconnect(self._get_radial_range_1D)
        self.ui.radial_high_1D.textChanged.disconnect(self._get_radial_range_1D)
        self.ui.radial_autoRange_1D.stateChanged.disconnect(self._get_radial_range_1D)

    def _connect_azim_range_1D_signals(self):
        """Connect signals for azimuth 1D range"""
        self.ui.azim_low_1D.textChanged.connect(self._get_azim_range_1D)
        self.ui.azim_high_1D.textChanged.connect(self._get_azim_range_1D)
        self.ui.azim_autoRange_1D.stateChanged.connect(self._get_azim_range_1D)

    def _disconnect_azim_range_1D_signals(self):
        """Disconnect signals for azimuth 1D range"""
        self.ui.azim_low_1D.textChanged.disconnect(self._get_azim_range_1D)
        self.ui.azim_high_1D.textChanged.disconnect(self._get_azim_range_1D)
        self.ui.azim_autoRange_1D.stateChanged.disconnect(self._get_azim_range_1D)

    def _connect_radial_range_2D_signals(self):
        """Connect signals for radial 2D range"""
        self.ui.radial_low_2D.textChanged.connect(self._get_radial_range_2D)
        self.ui.radial_high_2D.textChanged.connect(self._get_radial_range_2D)
        self.ui.radial_autoRange_2D.stateChanged.connect(self._get_radial_range_2D)

    def _disconnect_radial_range_2D_signals(self):
        """Disconnect signals for radial 2D range"""
        self.ui.radial_low_2D.textChanged.disconnect(self._get_radial_range_2D)
        self.ui.radial_high_2D.textChanged.disconnect(self._get_radial_range_2D)
        self.ui.radial_autoRange_2D.stateChanged.disconnect(self._get_radial_range_2D)

    def _connect_azim_range_2D_signals(self):
        """Connect signals for azimuth 2D range"""
        self.ui.azim_low_2D.textChanged.connect(self._get_azim_range_2D)
        self.ui.azim_high_2D.textChanged.connect(self._get_azim_range_2D)
        self.ui.azim_autoRange_2D.stateChanged.connect(self._get_azim_range_2D)

    def _disconnect_azim_range_2D_signals(self):
        """Disconnect signals for azimuth 2D range"""
        self.ui.azim_low_2D.textChanged.disconnect(self._get_azim_range_2D)
        self.ui.azim_high_2D.textChanged.disconnect(self._get_azim_range_2D)
        self.ui.azim_autoRange_2D.stateChanged.disconnect(self._get_azim_range_2D)

    def bai_1d(self, q):
        """Uses the integrator_thread attribute to call bai_1d
        """
        with self.integrator_thread.lock:
            if len(self.sphere.arches.index) > 0:
                self.integrator_thread.method = 'bai_1d_all'
            # # if self.ui.all1D.isChecked() or type(self.arch.idx) != int:
            # if self.ui.all1D.isChecked() or ('Overall' in self.arch_ids):
            #     self.integrator_thread.method = 'bai_1d_all'
            # else:
            #     self.integrator_thread.method = 'bai_1d_SI'
        self.data_1d.clear()
        self.setEnabled(False)
        self.integrator_thread.start()

    def bai_2d(self, q):
        """Uses the integrator_thread attribute to call bai_2d
        """
        with self.integrator_thread.lock:
            if len(self.sphere.arches.index) > 0:
                self.integrator_thread.method = 'bai_2d_all'
            # if self.ui.all2D.isChecked() or type(self.arch.idx) != int:
            # if self.ui.all2D.isChecked() or ('Overall' in self.arch_ids):
            #     self.integrator_thread.method = 'bai_2d_all'
            # else:
            #     self.integrator_thread.method = 'bai_2d_SI'
        self.data_2d.clear()
        self.setEnabled(False)
        self.integrator_thread.start()

    @staticmethod
    def run_pyfai_calib():
        # pyFAI_calib2_main()
        # launch(f'{current_directory}/pyFAI-calib2-xdart')
        # if getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS'):
        #     current_directory = sys._MEIPASS
        #     launch(f'{current_directory}/pyFAI-calib2')
        # else:
        #     pyFAI_calib2_main()
        process = subprocess.run(['pyFAI-calib2'], check=True, shell=True,
                                 stdout=subprocess.PIPE, universal_newlines=True)
        _ = process.stdout

    # @staticmethod
    def run_pyfai_drawmask(self):
        filters = f'Images (*.tif *tiff)'
        processFile, _ = QFileDialog().getOpenFileName(
            filter=filters,
            caption='Choose Image File',
            options=QFileDialog.DontUseNativeDialog
        )
        if not os.path.exists(processFile):
            print('No Image Chosen')
            return

        self.mask_window = MaskImageWidgetXdart()
        self.mask_window.setWindowModality(Qt.QtCore.Qt.WindowModal)
        self.mask_window.show()

        image = fabio.open(processFile).data
        pyFAI_drawmask_main(self.mask_window, image, processFile)
        # pyFAI_drawmask_main(self.mask_window, processFile)

        # mask = self.mask_window.getSelectionMask()
        # postProcessId21([processFile], mask)

    def set_image_units(self):
        """Disable/Enable Qz-Qxy option if we are/are not in GI mode"""
        if not self.sphere.gi:
            self.ui.axis2D.removeItem(1)
        else:
            if self.ui.axis2D.count() == 1:
                self.ui.axis2D.addItem(_translate("Form", 'Qz-Qxy'))

    def _update_axes(self, n):
        """Updates axes to allow user to set Qz-Qxy integration ranges in GI mode"""
        pass
        # if n == 1:
        #     if self.ui.unit_2D


class advancedParameters(Qt.QtWidgets.QWidget):
    """Pop up window for setting more advanced integration parameters.

    attributes:
        name: str, name of the window
        parameter: pyqtgraph Parameter, parameters displayed
        tree: pyqtgraph ParameterTree, tree to hold parameter
        layout: QVBoxLayout, holds tree

    methods:
        process_change: Handles sigTreeStateChanged signal from tree

    signals:
        sigUpdateArgs: str, sends own name for updating args.
    """
    sigUpdateArgs = Qt.QtCore.Signal(str)

    def __init__(self, parameter, name, parent=None):
        #ic()
        super().__init__(parent)
        self.name = name
        self.parameter = parameter
        self.tree = pg.parametertree.ParameterTree()
        self.tree.addParameters(parameter)
        self.parameter.sigTreeStateChanged.connect(self.process_change)
        self.layout = Qt.QtWidgets.QVBoxLayout(self)
        self.setLayout(self.layout)
        self.layout.addWidget(self.tree)

    def process_change(self, tree, changes):
        #ic()
        self.sigUpdateArgs.emit(self.name)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'c:\Users\walroth\Documents\repos\xdart\xdart\experiments\ttheta_scan\gui\integratorUI.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(218, 264)
        self.verticalLayout = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.frame1D = QtWidgets.QFrame(Form)
        self.frame1D.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame1D.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame1D.setObjectName("frame1D")
        self.layout1D = QtWidgets.QVBoxLayout(self.frame1D)
        self.layout1D.setContentsMargins(0, 0, 0, 0)
        self.layout1D.setSpacing(0)
        self.layout1D.setObjectName("layout1D")
        self.label = QtWidgets.QLabel(self.frame1D)
        font = QtGui.QFont()
        font.setPointSize(10)
        font.setBold(True)
        font.setWeight(75)
        self.label.setFont(font)
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        self.label.setObjectName("label")
        self.layout1D.addWidget(self.label)
        self.frame_3 = QtWidgets.QFrame(self.frame1D)
        self.frame_3.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_3.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_3.setLineWidth(0)
        self.frame_3.setObjectName("frame_3")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.frame_3)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setSpacing(2)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.integrate1D = QtWidgets.QPushButton(self.frame_3)
        self.integrate1D.setObjectName("integrate1D")
        self.horizontalLayout.addWidget(self.integrate1D)
        self.advanced1D = QtWidgets.QPushButton(self.frame_3)
        self.advanced1D.setObjectName("advanced1D")
        self.horizontalLayout.addWidget(self.advanced1D)
        self.all1D = QtWidgets.QCheckBox(self.frame_3)
        self.all1D.setObjectName("all1D")
        self.horizontalLayout.addWidget(self.all1D)
        self.layout1D.addWidget(self.frame_3)
        self.verticalLayout.addWidget(self.frame1D)
        self.frame2D = QtWidgets.QFrame(Form)
        self.frame2D.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame2D.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame2D.setObjectName("frame2D")
        self.layout2D = QtWidgets.QVBoxLayout(self.frame2D)
        self.layout2D.setContentsMargins(0, 0, 0, 0)
        self.layout2D.setSpacing(0)
        self.layout2D.setObjectName("layout2D")
        self.label_2 = QtWidgets.QLabel(self.frame2D)
        font = QtGui.QFont()
        font.setPointSize(10)
        font.setBold(True)
        font.setWeight(75)
        self.label_2.setFont(font)
        self.label_2.setAlignment(QtCore.Qt.AlignCenter)
        self.label_2.setObjectName("label_2")
        self.layout2D.addWidget(self.label_2)
        self.frame_4 = QtWidgets.QFrame(self.frame2D)
        self.frame_4.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame_4.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_4.setObjectName("frame_4")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.frame_4)
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_2.setSpacing(2)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.integrate2D = QtWidgets.QPushButton(self.frame_4)
        self.integrate2D.setObjectName("integrate2D")
        self.horizontalLayout_2.addWidget(self.integrate2D)
        self.advanced2D = QtWidgets.QPushButton(self.frame_4)
        self.advanced2D.setObjectName("advanced2D")
        self.horizontalLayout_2.addWidget(self.advanced2D)
        self.all2D = QtWidgets.QCheckBox(self.frame_4)
        self.all2D.setObjectName("all2D")
        self.horizontalLayout_2.addWidget(self.all2D)
        self.layout2D.addWidget(self.frame_4)
        self.verticalLayout.addWidget(self.frame2D)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label.setText(_translate("Form", "1D"))
        self.integrate1D.setText(_translate("Form", "Integrate"))
        self.advanced1D.setText(_translate("Form", "Advanced..."))
        self.all1D.setText(_translate("Form", "All"))
        self.label_2.setText(_translate("Form", "2D"))
        self.integrate2D.setText(_translate("Form", "Integrate"))
        self.advanced2D.setText(_translate("Form", "Advanced..."))
        self.all2D.setText(_translate("Form", "All"))

